import os
from typing import Any

import trafaret as t

from ai.backend.common import config
from ai.backend.common import validators as tx
from ai.backend.common.types import ResourceGroupType

from .affinity_map import AffinityPolicy
from .stats import StatModes
from .types import AgentBackend

coredump_defaults = {
    "enabled": False,
    "path": "./coredumps",
    "backup-count": 10,
    "size-limit": "64M",
}

default_sync_container_lifecycles_config = {
    "enabled": True,
    "interval": 10.0,
}

_default_pyroscope_config: dict[str, Any] = {
    "enabled": False,
    "app-name": None,
    "server-addr": None,
    "sample-rate": None,
}

agent_local_config_iv = (
    t.Dict({
        t.Key("agent"): t.Dict({
            tx.AliasedKey(["backend", "mode"]): tx.Enum(AgentBackend),
            t.Key("rpc-listen-addr", default=("", 6001)): tx.HostPortPair(allow_blank_host=True),
            t.Key("service-addr", default=("0.0.0.0", 6003)): tx.HostPortPair,
            t.Key("ssl-enabled", default=False): t.Bool,
            t.Key("ssl-cert", default=None): t.Null | tx.Path(type="file"),
            t.Key("ssl-key", default=None): t.Null | tx.Path(type="file"),
            t.Key("advertised-rpc-addr", default=None): t.Null | tx.HostPortPair,
            t.Key("rpc-auth-manager-public-key", default=None): t.Null | tx.Path(type="file"),
            t.Key("rpc-auth-agent-keypair", default=None): t.Null | tx.Path(type="file"),
            t.Key("agent-sock-port", default=6007): t.ToInt[1024:65535],
            t.Key("id", default=None): t.Null | t.String,
            t.Key("ipc-base-path", default="/tmp/backend.ai/ipc"): tx.Path(
                type="dir", auto_create=True
            ),
            t.Key("var-base-path", default="./var/lib/backend.ai"): tx.Path(
                type="dir", auto_create=True
            ),
            t.Key("mount-path", default=None): t.Null | tx.Path(type="dir", auto_create=True),
            t.Key("cohabiting-storage-proxy", default=True): t.Bool(),
            t.Key("public-host", default=None): t.Null | t.String,
            t.Key("region", default=None): t.Null | t.String,
            t.Key("instance-type", default=None): t.Null | t.String,
            t.Key("scaling-group", default="default"): t.String,
            t.Key("scaling-group-type", default=ResourceGroupType.COMPUTE): t.Enum(
                *(e.value for e in ResourceGroupType)
            ),
            t.Key("pid-file", default=os.devnull): tx.Path(
                type="file", allow_nonexisting=True, allow_devnull=True
            ),
            t.Key("event-loop", default="asyncio"): t.Enum("asyncio", "uvloop"),
            t.Key("skip-manager-detection", default=False): t.ToBool,
            tx.AliasedKey(["aiomonitor-termui-port", "aiomonitor-port"], default=38200): t.ToInt[
                1:65535
            ],
            t.Key("aiomonitor-webui-port", default=39200): t.ToInt[1:65535],
            t.Key("metadata-server-bind-host", default="0.0.0.0"): t.String,
            t.Key("metadata-server-port", default=40128): t.ToInt[1:65535],
            t.Key("allow-compute-plugins", default=None): t.Null | tx.ToSet,
            t.Key("block-compute-plugins", default=None): t.Null | tx.ToSet,
            t.Key("allow-network-plugins", default=None): t.Null | tx.ToSet,
            t.Key("block-network-plugins", default=None): t.Null | tx.ToSet,
            t.Key("image-commit-path", default="./tmp/backend.ai/commit"): tx.Path(
                type="dir", auto_create=True
            ),
            t.Key("abuse-report-path", default=None): t.Null
            | tx.Path(type="dir", allow_nonexisting=True),
            t.Key("force-terminate-abusing-containers", default=False): t.ToBool,
            t.Key("kernel-creation-concurrency", default=4): t.ToInt[1:32],
            t.Key("use-experimental-redis-event-dispatcher", default=False): t.ToBool,
            t.Key(
                "sync-container-lifecycles", default=default_sync_container_lifecycles_config
            ): t.Dict({
                t.Key(
                    "enabled", default=default_sync_container_lifecycles_config["enabled"]
                ): t.ToBool,
                t.Key(
                    "interval", default=default_sync_container_lifecycles_config["interval"]
                ): t.ToFloat[0:],
            }).allow_extra("*"),
        }).allow_extra("*"),
        t.Key("container"): t.Dict({
            t.Key("kernel-uid", default=-1): tx.UserID,
            t.Key("kernel-gid", default=-1): tx.GroupID,
            t.Key("bind-host", default=""): t.String(allow_blank=True),
            t.Key("advertised-host", default=None): t.Null | t.String(),
            t.Key("port-range", default=(30000, 31000)): tx.PortRange,
            t.Key("stats-type", default=StatModes.DOCKER): t.Null
            | t.Enum(*(e.value for e in StatModes)),
            t.Key("sandbox-type", default="docker"): t.Enum("docker", "jail"),
            t.Key("jail-args", default=[]): t.List(t.String),
            t.Key("scratch-type"): t.Enum("hostdir", "hostfile", "memory", "k8s-nfs"),
            t.Key("scratch-root", default="./scratches"): tx.Path(type="dir", auto_create=True),
            t.Key("scratch-size", default="0"): tx.BinarySize,
            t.Key("scratch-nfs-address", default=None): t.Null | t.String,
            t.Key("scratch-nfs-options", default=None): t.Null | t.String,
            t.Key("alternative-bridge", default=None): t.Null | t.String,
        }).allow_extra("*"),
        t.Key("pyroscope", default=_default_pyroscope_config): t.Dict({
            t.Key("enabled", default=_default_pyroscope_config["enabled"]): t.ToBool,
            t.Key("app-name", default=_default_pyroscope_config["app-name"]): t.Null | t.String,
            t.Key("server-addr", default=_default_pyroscope_config["server-addr"]): t.Null
            | t.String,
            t.Key("sample-rate", default=_default_pyroscope_config["sample-rate"]): t.Null
            | t.ToInt[1:],
        }).allow_extra("*"),
        t.Key("logging"): t.Any,  # checked in ai.backend.logging
        t.Key("resource"): t.Dict({
            t.Key("reserved-cpu", default=1): t.Int,
            t.Key("reserved-mem", default="1G"): tx.BinarySize,
            t.Key("reserved-disk", default="8G"): tx.BinarySize,
            t.Key("allocation-order", default=["cuda", "rocm", "tpu", "cpu", "mem"]): t.List(
                t.String
            ),
            t.Key("affinity-policy", default=AffinityPolicy.INTERLEAVED.name): tx.Enum(
                AffinityPolicy,
                use_name=True,
            ),
        }).allow_extra("*"),
        t.Key("debug"): t.Dict({
            t.Key("enabled", default=False): t.ToBool,
            t.Key("asyncio", default=False): t.ToBool,
            t.Key("kernel-runner", default=False): t.ToBool,
            t.Key("enhanced-aiomonitor-task-info", default=False): t.ToBool,
            t.Key("skip-container-deletion", default=False): t.ToBool,
            t.Key("log-stats", default=False): t.ToBool,
            t.Key("log-kernel-config", default=False): t.ToBool,
            t.Key("log-alloc-map", default=False): t.ToBool,
            t.Key("log-events", default=False): t.ToBool,
            t.Key("log-heartbeats", default=False): t.ToBool,
            t.Key("heartbeat-interval", default=20.0): t.Float,
            t.Key("log-docker-events", default=False): t.ToBool,
            t.Key("coredump", default=coredump_defaults): t.Dict({
                t.Key("enabled", default=coredump_defaults["enabled"]): t.ToBool,
                t.Key("path", default=coredump_defaults["path"]): tx.Path(
                    type="dir", auto_create=True
                ),
                t.Key("backup-count", default=coredump_defaults["backup-count"]): t.Int[1:],
                t.Key("size-limit", default=coredump_defaults["size-limit"]): tx.BinarySize,
            }).allow_extra("*"),
        }).allow_extra("*"),
    })
    .merge(config.etcd_config_iv)
    .allow_extra("*")
)

docker_extra_config_iv = t.Dict({
    t.Key("container"): t.Dict({
        t.Key("swarm-enabled", default=False): t.ToBool,
    }).allow_extra("*"),
}).allow_extra("*")

default_container_logs_config = {
    "max-length": "10M",  # the maximum tail size
    "chunk-size": "64K",  # used when storing logs to Redis as a side-channel to the event bus
}

DEFAULT_PULL_TIMEOUT = 2 * 60 * 60  # 2 hours
DEFAULT_PUSH_TIMEOUT = None  # Infinite
DEFAULT_KERNEL_INIT_POLLING_ATTEMPT = 10
DEFAULT_KERNEL_INIT_POLLING_TIMEOUT = 60.0  # 60 seconds
DEFAULT_KERNEL_INIT_TIMEOUT = 60.0  # 60 seconds

default_api_config = {"pull-timeout": DEFAULT_PULL_TIMEOUT, "push-timeout": DEFAULT_PUSH_TIMEOUT}
default_kernel_lifecycles = {
    "init-polling-attempt": DEFAULT_KERNEL_INIT_POLLING_ATTEMPT,
    "init-polling-timeout-sec": DEFAULT_KERNEL_INIT_POLLING_TIMEOUT,
    "init-timeout-sec": DEFAULT_KERNEL_INIT_TIMEOUT,
}


agent_etcd_config_iv = t.Dict({
    t.Key("container-logs", default=default_container_logs_config): t.Dict({
        t.Key("max-length", default=default_container_logs_config["max-length"]): tx.BinarySize(),
        t.Key("chunk-size", default=default_container_logs_config["chunk-size"]): tx.BinarySize(),
    }).allow_extra("*"),
    t.Key("api", default=default_api_config): t.Dict({
        t.Key("pull-timeout", default=default_api_config["pull-timeout"]): tx.ToNone
        | t.ToFloat[0:],  # Set the image pull timeout in seconds
        t.Key("push-timeout", default=default_api_config["push-timeout"]): tx.ToNone
        | t.ToFloat[
            0:
        ],  # Set the image push timeout in seconds. 'None' indicates an infinite timeout.
    }).allow_extra("*"),
    t.Key("kernel-lifecycles", default=default_kernel_lifecycles): t.Dict({
        t.Key(
            "init-polling-attempt",
            default=default_kernel_lifecycles["init-polling-attempt"],
        ): t.ToInt,
        t.Key(
            "init-polling-timeout-sec",
            default=default_kernel_lifecycles["init-polling-timeout-sec"],
        ): t.ToFloat,
        t.Key(
            "init-timeout-sec",
            default=default_kernel_lifecycles["init-timeout-sec"],
        ): t.ToFloat,
    }),
}).allow_extra("*")

container_etcd_config_iv = t.Dict({
    t.Key("kernel-uid", optional=True): t.ToInt,
    t.Key("kernel-gid", optional=True): t.ToInt,
}).allow_extra("*")



from __future__ import annotations

import asyncio
import os
from collections import OrderedDict
from typing import Any, Dict, FrozenSet, Mapping, Sequence, override

from ai.backend.common.docker import ImageRef
from ai.backend.common.events import EventProducer
from ai.backend.common.types import AgentId, CommitStatus, KernelId, SessionId

from ..kernel import AbstractCodeRunner, AbstractKernel, NextResult, ResultRecord
from ..resources import KernelResourceSpec
from ..types import AgentEventData


class DummyKernel(AbstractKernel):
    dummy_config: Mapping[str, Any]

    def __init__(
        self,
        kernel_id: KernelId,
        session_id: SessionId,
        agent_id: AgentId,
        network_id: str,
        image: ImageRef,
        version: int,
        *,
        agent_config: Mapping[str, Any],
        resource_spec: KernelResourceSpec,
        service_ports: Any,  # TODO: type-annotation
        environ: Mapping[str, Any],
        data: Dict[str, Any],
        dummy_config: Mapping[str, Any],
    ) -> None:
        super().__init__(
            kernel_id,
            session_id,
            agent_id,
            network_id,
            image,
            version,
            agent_config=agent_config,
            resource_spec=resource_spec,
            service_ports=service_ports,
            data=data,
            environ=environ,
        )
        self.is_commiting = False
        self.dummy_config = dummy_config
        self.dummy_kernel_cfg = self.dummy_config["kernel"]

    async def close(self) -> None:
        pass

    async def create_code_runner(
        self,
        event_producer: EventProducer,
        *,
        client_features: FrozenSet[str],
        api_version: int,
    ) -> "AbstractCodeRunner":
        if self.dummy_kernel_cfg["use-fake-code-runner"]:
            return await DummyFakeCodeRunner.new(
                self.kernel_id,
                self.session_id,
                event_producer,
                kernel_host=self.data["kernel_host"],
                repl_in_port=self.data["repl_in_port"],
                repl_out_port=self.data["repl_out_port"],
                exec_timeout=0,
                client_features=client_features,
            )
        else:
            return await DummyCodeRunner.new(
                self.kernel_id,
                self.session_id,
                event_producer,
                kernel_host=self.data["kernel_host"],
                repl_in_port=self.data["repl_in_port"],
                repl_out_port=self.data["repl_out_port"],
                exec_timeout=0,
                client_features=client_features,
            )

    async def check_status(self):
        delay = self.dummy_kernel_cfg["delay"]["check-status"]
        await asyncio.sleep(delay)
        return {}

    async def get_completions(self, text, opts):
        delay = self.dummy_kernel_cfg["delay"]["get-completions"]
        await asyncio.sleep(delay)
        return {"status": "finished", "completions": []}

    async def get_logs(self):
        delay = self.dummy_kernel_cfg["delay"]["get-logs"]
        await asyncio.sleep(delay)
        return {"logs": "my logs"}

    async def interrupt_kernel(self):
        delay = self.dummy_kernel_cfg["delay"]["interrupt-kernel"]
        await asyncio.sleep(delay)

    async def start_service(self, service, opts):
        delay = self.dummy_kernel_cfg["delay"]["start-service"]
        await asyncio.sleep(delay)

    async def start_model_service(self, model_service: Mapping[str, Any]):
        delay = self.dummy_kernel_cfg["delay"]["start-model-service"]
        await asyncio.sleep(delay)
        return {}

    async def shutdown_service(self, service):
        delay = self.dummy_kernel_cfg["delay"]["shutdown-service"]
        await asyncio.sleep(delay)

    async def check_duplicate_commit(self, kernel_id, subdir) -> CommitStatus:
        if self.is_commiting:
            return CommitStatus.ONGOING
        return CommitStatus.READY

    async def commit(
        self,
        kernel_id,
        subdir,
        *,
        canonical: str | None = None,
        filename: str | None = None,
        extra_labels: dict[str, str] = {},
    ) -> None:
        self.is_commiting = True
        delay = self.dummy_kernel_cfg["delay"]["commit"]
        await asyncio.sleep(delay)
        self.is_commiting = False

    async def get_service_apps(self):
        delay = self.dummy_kernel_cfg["delay"]["get-service-apps"]
        await asyncio.sleep(delay)
        return {
            "status": "done",
            "data": [],
        }

    @override
    async def accept_file(self, container_path: os.PathLike | str, filedata: bytes) -> None:
        delay = self.dummy_kernel_cfg["delay"]["accept-file"]
        await asyncio.sleep(delay)

    @override
    async def download_file(self, container_path: os.PathLike | str) -> bytes:
        delay = self.dummy_kernel_cfg["delay"]["download-file"]
        await asyncio.sleep(delay)
        return b""

    @override
    async def download_single(self, container_path: os.PathLike | str) -> bytes:
        delay = self.dummy_kernel_cfg["delay"]["download-single"]
        await asyncio.sleep(delay)
        return b""

    @override
    async def list_files(self, container_path: os.PathLike | str):
        delay = self.dummy_kernel_cfg["delay"]["list-files"]
        await asyncio.sleep(delay)
        return {"files": "", "errors": "", "abspath": ""}

    async def notify_event(self, evdata: AgentEventData):
        raise NotImplementedError


class DummyCodeRunner(AbstractCodeRunner):
    kernel_host: str
    repl_in_port: int
    repl_out_port: int

    def __init__(
        self,
        kernel_id,
        session_id,
        event_producer,
        *,
        kernel_host,
        repl_in_port,
        repl_out_port,
        exec_timeout=0,
        client_features=None,
    ) -> None:
        super().__init__(
            kernel_id,
            session_id,
            event_producer,
            exec_timeout=exec_timeout,
            client_features=client_features,
        )
        self.kernel_host = kernel_host
        self.repl_in_port = repl_in_port
        self.repl_out_port = repl_out_port

    async def get_repl_in_addr(self) -> str:
        return f"tcp://{self.kernel_host}:{self.repl_in_port}"

    async def get_repl_out_addr(self) -> str:
        return f"tcp://{self.kernel_host}:{self.repl_out_port}"


class DummyFakeCodeRunner(AbstractCodeRunner):
    kernel_host: str
    repl_in_port: int
    repl_out_port: int

    input_sock: None  # type: ignore[assignment]
    output_sock: None  # type: ignore[assignment]
    zctx: None  # type: ignore[assignment]

    def __init__(
        self,
        kernel_id,
        session_id,
        event_producer,
        *,
        kernel_host,
        repl_in_port,
        repl_out_port,
        exec_timeout=0,
        client_features=None,
    ) -> None:
        self.zctx = None
        self.input_sock = None
        self.output_sock = None

        self.completion_queue = asyncio.Queue(maxsize=128)
        self.service_queue = asyncio.Queue(maxsize=128)
        self.model_service_queue = asyncio.Queue(maxsize=128)
        self.service_apps_info_queue = asyncio.Queue(maxsize=128)
        self.status_queue = asyncio.Queue(maxsize=128)
        self.output_queue = None
        self.pending_queues = OrderedDict()
        self.current_run_id = None
        self.read_task = None
        self.status_task = None
        self.watchdog_task = None
        self._closed = False

        self.kernel_host = kernel_host
        self.repl_in_port = repl_in_port
        self.repl_out_port = repl_out_port

        self.event_producer = event_producer

    async def __ainit__(self) -> None:
        return

    def __setstate__(self, props):
        self.__dict__.update(props)
        self.zctx = None
        self.input_sock = None
        self.output_sock = None

        self.completion_queue = asyncio.Queue(maxsize=128)
        self.service_queue = asyncio.Queue(maxsize=128)
        self.model_service_queue = asyncio.Queue(maxsize=128)
        self.service_apps_info_queue = asyncio.Queue(maxsize=128)
        self.status_queue = asyncio.Queue(maxsize=128)
        self.output_queue = None
        self.pending_queues = OrderedDict()
        self.current_run_id = None
        self.read_task = None
        self.status_task = None
        self.watchdog_task = None
        self._closed = False

    async def get_repl_in_addr(self) -> str:
        return f"tcp://{self.kernel_host}:{self.repl_in_port}"

    async def get_repl_out_addr(self) -> str:
        return f"tcp://{self.kernel_host}:{self.repl_out_port}"

    async def close(self) -> None:
        return None

    async def ping_status(self):
        return None

    async def feed_batch(self, opts):
        return None

    async def feed_code(self, text: str):
        return None

    async def feed_input(self, text: str):
        return None

    async def feed_interrupt(self):
        return None

    async def feed_and_get_status(self):
        return None

    async def feed_and_get_completion(self, code_text, opts):
        return []

    async def feed_start_model_service(self, model_info):
        return {"status": "failed", "error": "not-implemented"}

    async def feed_start_service(self, service_info):
        return {"status": "failed", "error": "not-implemented"}

    async def feed_service_apps(self):
        return {"status": "failed", "error": "not-implemented"}

    @staticmethod
    def aggregate_console(
        result: NextResult, records: Sequence[ResultRecord], api_ver: int
    ) -> None:
        return

    async def get_next_result(self, api_ver=2, flush_timeout=2.0) -> NextResult:
        return {
            "runId": self.current_run_id,
            "status": "finished",
            "exitCode": None,
            "options": None,
        }

    async def attach_output_queue(self, run_id: str | None) -> None:
        return

    def resume_output_queue(self) -> None:
        return

    def next_output_queue(self) -> None:
        return

    async def read_output(self) -> None:
        return


async def prepare_krunner_env(local_config: Mapping[str, Any]) -> Mapping[str, Sequence[str]]:
    return {}



from __future__ import annotations

import enum
from collections import defaultdict
from typing import TYPE_CHECKING, Optional, Sequence

import attr
import networkx as nx

from ai.backend.common.types import DeviceName

if TYPE_CHECKING:
    from .resources import AbstractComputeDevice


class AffinityPolicy(enum.Enum):
    PREFER_SINGLE_NODE = 0
    INTERLEAVED = 1


@attr.define()
class AffinityHint:
    devices: Optional[Sequence[AbstractComputeDevice]]
    affinity_map: AffinityMap
    policy: AffinityPolicy


class AffinityMap(nx.Graph):
    """
    Represents the NUMA distance matrix of all device pairs from all compute device plugins.
    """

    def __init__(self) -> None:
        self.max_weight = 0
        super().__init__()

    def get_largest_device_cluster_with_lowest_distance_from_src_device(
        self,
        device_name: DeviceName,
        src_device: AbstractComputeDevice,
    ) -> Sequence[Sequence[AbstractComputeDevice]]:
        distance_sets: dict[int, nx.Graph] = defaultdict(nx.Graph)
        for v in self.neighbors(src_device):
            if v.device_name == device_name:
                weight = self.edges[src_device, v]["weight"]
                distance_sets[weight].add_edge(src_device, v, weight=weight)
        device_cluster_list = []
        for distance, device_set in distance_sets.items():
            components = nx.connected_components(device_set)
            for component in components:
                device_cluster_list.append((distance, component))
        # sort by: low distance first, large component first
        device_cluster_list.sort(key=lambda item: (item[0], -len(item[1])))
        largest_components: list[list[AbstractComputeDevice]] = []
        for distance, device_set in device_cluster_list:
            largest_component: list[AbstractComputeDevice] = []
            for device in device_set:
                if device == src_device:
                    continue
                largest_component.append(device)
            largest_components.append(largest_component)
        return largest_components

    def get_device_clusters_with_lowest_distance(
        self,
        device_name: DeviceName,
    ) -> Sequence[Sequence[AbstractComputeDevice]]:
        device_cluster_list = []
        # FIXME: this is the intended logic but causes infinite loop.
        # for weight in range(self.max_weight + 1):
        for weight in [0]:
            subgraph = nx.subgraph_view(
                self,
                filter_node=lambda u: u.device_name == device_name,
                filter_edge=lambda u, v: self.edges[u, v]["weight"] == weight,
            )
            components = nx.connected_components(subgraph)
            for component in components:
                device_cluster_list.append((weight, component))
        # sort by: low distance first, large component first
        device_cluster_list.sort(key=lambda item: (item[0], -len(item[1])))
        return [device_set for distance, device_set in device_cluster_list]

    def get_distance_ordered_neighbors(
        self,
        src_devices: Optional[Sequence[AbstractComputeDevice]],
        device_name: DeviceName,
    ) -> tuple[Sequence[Sequence[AbstractComputeDevice]], Sequence[AbstractComputeDevice]]:
        """
        Get the list of neighbor device clusters and their distance from the given source_devices
        with the same name.

        Example:
            Given a 4-core dual socket system with two GPUs per socket:

            If the prior allocator has assigned (gpu0@node0, gpu2@node1),
            it will return:
               primary: cpu0-3@node0, cpu4-7@node1 [interleaved]
               secondary: (empty) [fill-remaining]

            If the prior allocator has assigned (gpu0@node0, gpu1@node0),
            it will return:
              primary: cpu0-3@node0 [interleaved]
              secondary: cpu4-7@node1 [fill-remaining]

        If source_devices is None, it will return the first largest connected component from the
        device distance matrix sharing the lowest distance values.
        """
        if src_devices:
            assert next(iter(src_devices)).device_name != device_name
            src_numa_nodes = set(src_device.numa_node for src_device in src_devices)
            primary_sets = []
            secondary_set = set()
            for numa_node in src_numa_nodes:
                primary_set = set()
                for u in self.nodes:
                    if u.device_name == device_name:
                        if u.numa_node == numa_node:
                            primary_set.add(u)
                        else:
                            secondary_set.add(u)
                primary_sets.append(primary_set)
            for primary_set in primary_sets:
                secondary_set -= primary_set
            return [*(list(primary_set) for primary_set in primary_sets)], list(secondary_set)
        else:
            components = self.get_device_clusters_with_lowest_distance(device_name)
            return components, []

    @classmethod
    def build(cls, devices: Sequence[AbstractComputeDevice]) -> AffinityMap:
        # TODO: allow compute plugins to customize distance calculation
        g = cls()
        max_weight = 0
        devices_copy = list(devices)
        while devices_copy:
            device1 = devices_copy.pop(0)
            g.add_edge(device1, device1, weight=0)
            for device2 in devices_copy:
                # weight = abs((max(0, device2.numa_node or 0)) - max(0, (device1.numa_node or 0)))
                weight = 0 if device1.numa_node == device2.numa_node else 1
                if max_weight < weight:
                    max_weight = weight
                g.add_edge(device1, device2, weight=weight)
        g.max_weight = max_weight
        return g



from dataclasses import dataclass
from typing import (
    NewType,
)

from ai.backend.common.types import AgentId, DeviceId, KernelId, MetricKey

MetricValueFieldKey = NewType("MetricValueFieldKey", str)
MetricValueFieldPair = tuple[MetricValueFieldKey, str]

CURRENT_METRIC_KEY = MetricValueFieldKey("current")
CAPACITY_METRIC_KEY = MetricValueFieldKey("capacity")
PCT_METRIC_KEY = MetricValueFieldKey("pct")


@dataclass
class FlattenedKernelMetric:
    agent_id: AgentId
    kernel_id: KernelId
    key: MetricKey
    value_pairs: list[MetricValueFieldPair]


@dataclass
class FlattenedDeviceMetric:
    agent_id: AgentId
    device_id: DeviceId
    key: MetricKey
    value_pairs: list[MetricValueFieldPair]



from __future__ import annotations

import asyncio
import codecs
import io
import json
import logging
import math
import os
import re
import secrets
import time
from abc import ABCMeta, abstractmethod
from collections import OrderedDict, UserDict
from collections.abc import (
    Mapping,
    Sequence,
)
from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    FrozenSet,
    List,
    Literal,
    NotRequired,
    Optional,
    Set,
    Tuple,
    TypedDict,
    Union,
    cast,
)

import zmq
import zmq.asyncio
from async_timeout import timeout

from ai.backend.common import msgpack
from ai.backend.common.asyncio import current_loop
from ai.backend.common.docker import ImageRef
from ai.backend.common.enum_extension import StringSetFlag
from ai.backend.common.events import (
    EventProducer,
    KernelLifecycleEventReason,
    ModelServiceStatusEvent,
)
from ai.backend.common.types import (
    AgentId,
    CommitStatus,
    KernelId,
    ModelServiceStatus,
    ServicePort,
    SessionId,
    aobject,
)
from ai.backend.logging import BraceStyleAdapter

from .exception import InvalidSocket, UnsupportedBaseDistroError
from .resources import KernelResourceSpec
from .types import AgentEventData, KernelLifecycleStatus

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

# msg types visible to the API client.
# (excluding control signals such as 'finished' and 'waiting-input'
# since they are passed as separate status field.)
ConsoleItemType = Literal[
    "stdout",
    "stderr",
    "media",
    "html",
    "log",
    "completion",
]
outgoing_msg_types: FrozenSet[ConsoleItemType] = frozenset([
    "stdout",
    "stderr",
    "media",
    "html",
    "log",
    "completion",
])
ResultType = Union[
    ConsoleItemType,
    Literal[
        "continued",
        "clean-finished",
        "build-finished",
        "finished",
        "exec-timeout",
        "waiting-input",
    ],
]


class KernelFeatures(StringSetFlag):
    UID_MATCH = "uid-match"
    USER_INPUT = "user-input"
    BATCH_MODE = "batch"
    QUERY_MODE = "query"
    TTY_MODE = "tty"


class ClientFeatures(StringSetFlag):
    INPUT = "input"
    CONTINUATION = "continuation"


# TODO: use Python 3.7 contextvars for per-client feature selection
default_client_features = frozenset({
    ClientFeatures.INPUT.value,
    ClientFeatures.CONTINUATION.value,
})
default_api_version = 4


class RunEvent(Exception):
    data: Any

    def __init__(self, data=None):
        super().__init__()
        self.data = data


class InputRequestPending(RunEvent):
    pass


class CleanFinished(RunEvent):
    pass


class BuildFinished(RunEvent):
    pass


class RunFinished(RunEvent):
    pass


class ExecTimeout(RunEvent):
    pass


@dataclass
class ResultRecord:
    msg_type: ResultType
    data: Optional[str] = None


class NextResult(TypedDict):
    runId: Optional[str]
    status: ResultType
    exitCode: Optional[int]
    options: Optional[Mapping[str, Any]]
    # v1
    stdout: NotRequired[Optional[str]]
    stderr: NotRequired[Optional[str]]
    media: NotRequired[Sequence[Any]]
    html: NotRequired[Sequence[Any]]
    # v2
    console: NotRequired[Sequence[Any]]


class AbstractKernel(UserDict, aobject, metaclass=ABCMeta):
    version: int
    agent_config: Mapping[str, Any]
    session_id: SessionId
    kernel_id: KernelId
    agent_id: AgentId
    network_id: str
    container_id: Optional[str]
    image: ImageRef
    resource_spec: KernelResourceSpec
    service_ports: List[ServicePort]
    data: Dict[Any, Any]
    last_used: float
    termination_reason: Optional[KernelLifecycleEventReason]
    clean_event: Optional[asyncio.Future]
    stats_enabled: bool
    # FIXME: apply TypedDict to data in Python 3.8
    environ: Mapping[str, Any]
    status: KernelLifecycleStatus

    _tasks: Set[asyncio.Task]

    runner: Optional[AbstractCodeRunner]

    def __init__(
        self,
        kernel_id: KernelId,
        session_id: SessionId,
        agent_id: AgentId,
        network_id: str,
        image: ImageRef,
        version: int,
        *,
        agent_config: Mapping[str, Any],
        resource_spec: KernelResourceSpec,
        service_ports: Any,  # TODO: type-annotation
        data: Dict[Any, Any],
        environ: Mapping[str, Any],
    ) -> None:
        self.agent_config = agent_config
        self.kernel_id = kernel_id
        self.session_id = session_id
        self.agent_id = agent_id
        self.network_id = network_id
        self.image = image
        self.version = version
        self.resource_spec = resource_spec
        self.service_ports = service_ports
        self.data = data
        self.last_used = time.monotonic()
        self.termination_reason = None
        self.clean_event = None
        self.stats_enabled = False
        self._tasks = set()
        self.environ = environ
        self.runner = None
        self.container_id = None
        self.state = KernelLifecycleStatus.PREPARING

    async def init(self, event_producer: EventProducer) -> None:
        log.debug(
            "kernel.init(k:{0}, api-ver:{1}, client-features:{2}): starting new runner",
            self.kernel_id,
            default_api_version,
            default_client_features,
        )
        self.runner = await self.create_code_runner(
            event_producer, client_features=default_client_features, api_version=default_api_version
        )

    def __getstate__(self) -> Mapping[str, Any]:
        props = self.__dict__.copy()
        del props["agent_config"]
        del props["clean_event"]
        del props["_tasks"]
        return props

    def __setstate__(self, props) -> None:
        # Used when a `Kernel` object is loaded from pickle data.
        if "state" not in props:
            props["state"] = KernelLifecycleStatus.RUNNING
        self.__dict__.update(props)
        # agent_config is set by the pickle.loads() caller.
        self.clean_event = None
        self._tasks = set()

    @abstractmethod
    async def close(self) -> None:
        """
        Release internal resources used for interacting with the kernel.
        Note that this does NOT terminate the container.
        """
        pass

    # We don't have "allocate_slots()" method here because:
    # - resource_spec is initialized by allocating slots at computer's alloc_map
    #   when creating new kernels.
    # - restoration from running containers is done by computer's classmethod
    #   "restore_from_container"

    def release_slots(self, computer_ctxs) -> None:
        """
        Release the resource slots occupied by the kernel
        to the allocation maps.
        """
        for accel_key, accel_alloc in self.resource_spec.allocations.items():
            computer_ctxs[accel_key].alloc_map.free(accel_alloc)

    @abstractmethod
    async def create_code_runner(
        self,
        event_producer: EventProducer,
        *,
        client_features: FrozenSet[str],
        api_version: int,
    ) -> "AbstractCodeRunner":
        raise NotImplementedError

    @abstractmethod
    async def check_status(self):
        raise NotImplementedError

    @abstractmethod
    async def get_completions(self, text, opts):
        raise NotImplementedError

    @abstractmethod
    async def get_logs(self):
        raise NotImplementedError

    @abstractmethod
    async def interrupt_kernel(self):
        raise NotImplementedError

    @abstractmethod
    async def start_service(self, service, opts):
        raise NotImplementedError

    @abstractmethod
    async def start_model_service(self, model_service):
        raise NotImplementedError

    @abstractmethod
    async def shutdown_service(self, service):
        raise NotImplementedError

    @abstractmethod
    async def check_duplicate_commit(self, kernel_id, subdir) -> CommitStatus:
        raise NotImplementedError

    @abstractmethod
    async def commit(
        self,
        kernel_id,
        subdir,
        *,
        canonical: str | None = None,
        filename: str | None = None,
        extra_labels: dict[str, str] = {},
    ):
        raise NotImplementedError

    @abstractmethod
    async def get_service_apps(self):
        raise NotImplementedError

    @abstractmethod
    async def accept_file(self, container_path: os.PathLike | str, filedata) -> None:
        """
        Put the uploaded file to the designated container path.
        The path should be inside /home/work of the container.
        A relative path is interpreted as a subpath inside /home/work.

        WARNING: Since the implementations may use the scratch directory mounted as the home
        directory inside the container, the file may not be visible inside the container if the
        designated home-relative path overlaps with a vfolder mount.
        """
        raise NotImplementedError

    @abstractmethod
    async def download_file(self, container_path: os.PathLike | str) -> bytes:
        """
        Download the designated path (a single file or an entire directory) as a tar archive.
        The path should be inside /home/work of the container.
        A relative path is interpreted as a subpath inside /home/work.
        The return value is the raw byte stream of the archive itself, and it is the caller's
        responsibility to extract the tar archive.

        This API is intended to download a small set of files from the container filesystem.
        """
        raise NotImplementedError

    @abstractmethod
    async def download_single(self, container_path: os.PathLike | str) -> bytes:
        """
        Download the designated path (a single file) as a tar archive.
        The path should be inside /home/work of the container.
        A relative path is interpreted as a subpath inside /home/work.
        The return value is the content of the file *extracted* from the downloaded archive.

        This API is intended to download a small file from the container filesystem.
        """
        raise NotImplementedError

    @abstractmethod
    async def list_files(self, container_path: os.PathLike | str):
        """
        List the directory entries of the designated path.
        The path should be inside /home/work of the container.
        A relative path is interpreted as a subpath inside /home/work.
        """
        raise NotImplementedError

    @abstractmethod
    async def notify_event(self, evdata: AgentEventData):
        raise NotImplementedError

    async def ping(self) -> dict[str, float] | None:
        assert self.runner is not None
        return await self.runner.ping()

    async def execute(
        self,
        run_id: Optional[str],
        mode: Literal["batch", "query", "input", "continue"],
        text: str,
        *,
        opts: Mapping[str, Any],
        api_version: int,
        flush_timeout: float,
    ) -> NextResult:
        myself = asyncio.current_task()
        assert myself is not None
        assert self.runner is not None
        self._tasks.add(myself)
        try:
            await self.runner.attach_output_queue(run_id)
            try:
                if mode == "batch":
                    await self.runner.feed_batch(opts)
                elif mode == "query":
                    await self.runner.feed_code(text)
                elif mode == "input":
                    await self.runner.feed_input(text)
                elif mode == "continue":
                    pass
            except zmq.ZMQError:
                # cancel the operation by myself
                # since the peer is gone.
                raise asyncio.CancelledError
            return await self.runner.get_next_result(
                api_ver=api_version,
                flush_timeout=flush_timeout,
            )
        except asyncio.CancelledError:
            await self.runner.close()
            raise
        finally:
            self._tasks.remove(myself)


_zctx = None


class RobustSocket:
    _zctx: zmq.asyncio.Context
    _sock: zmq.asyncio.Socket
    _socket_type: int
    _addr: str

    def __init__(
        self,
        socket_type: int,
        addr: str,
    ) -> None:
        self._init_zctx()
        self._socket_type = socket_type
        self._addr = addr
        self._sock = self._zctx.socket(self._socket_type)
        self._sock.connect(self._addr)
        self._sock.setsockopt(zmq.LINGER, 50)

    @property
    def addr(self) -> str:
        return self._addr

    @property
    def socket(self) -> zmq.asyncio.Socket:
        return self._sock

    def close(self) -> None:
        try:
            self._sock.close()
        except zmq.ZMQError:
            pass

    def _init_zctx(self) -> None:
        global _zctx
        if _zctx is None:
            _zctx = zmq.asyncio.Context()
        self._zctx = _zctx

    def recreate_socket(self):
        self._init_zctx()
        self._sock = self._zctx.socket(self._socket_type)
        self._sock.connect(self._addr)
        self._sock.setsockopt(zmq.LINGER, 50)


class SocketPair:
    input_sock: RobustSocket
    output_sock: RobustSocket

    def __init__(self, input_sock: RobustSocket, output_sock: RobustSocket):
        self.input_sock = input_sock
        self.output_sock = output_sock

    async def send_multipart(self, msg_parts: Sequence[bytes]) -> None:
        try:
            await self.input_sock.socket.send_multipart(msg_parts)
        except zmq.ZMQError as e:
            if e.errno in (zmq.ENOTSOCK, zmq.ETERM):
                log.warning(
                    f"Socket invalid, recreating socket (addr: {self.input_sock.addr}, err: {repr(e)})"
                )
                self.input_sock.recreate_socket()
                self.output_sock.recreate_socket()
                await self.input_sock.socket.send_multipart(msg_parts)
            else:
                raise

    async def recv_multipart(self) -> list[bytes]:
        try:
            return await self.output_sock.socket.recv_multipart()
        except zmq.ZMQError as e:
            if e.errno in (zmq.ENOTSOCK, zmq.ETERM):
                log.exception(f"Socket invalid (addr: {self.output_sock.addr}, err: {repr(e)})")
                raise InvalidSocket
            else:
                raise

    def close(self) -> None:
        self.input_sock.close()
        self.output_sock.close()


class AbstractCodeRunner(aobject, metaclass=ABCMeta):
    kernel_id: KernelId
    session_id: SessionId
    started_at: float
    finished_at: Optional[float]
    exec_timeout: float
    max_record_size: int
    client_features: FrozenSet[str]

    event_producer: EventProducer

    _sockets: Optional[SocketPair]

    completion_queue: asyncio.Queue[bytes]
    service_queue: asyncio.Queue[bytes]
    model_service_queue: asyncio.Queue[bytes]
    service_apps_info_queue: asyncio.Queue[bytes]
    status_queue: asyncio.Queue[bytes]
    _is_socket_invalid: bool
    output_queue: Optional[asyncio.Queue[ResultRecord]]
    current_run_id: Optional[str]
    pending_queues: OrderedDict[str, Tuple[asyncio.Event, asyncio.Queue[ResultRecord]]]

    read_task: Optional[asyncio.Task]
    status_task: Optional[asyncio.Task]
    watchdog_task: Optional[asyncio.Task]

    _closed: bool

    def __init__(
        self,
        kernel_id: KernelId,
        session_id: SessionId,
        event_producer: EventProducer,
        *,
        exec_timeout: float = 0,
        client_features: Optional[FrozenSet[str]] = None,
    ) -> None:
        global _zctx
        self.kernel_id = kernel_id
        self.session_id = session_id
        self.event_producer = event_producer
        self.started_at = time.monotonic()
        self.finished_at = None
        if not math.isfinite(exec_timeout) or exec_timeout < 0:
            raise ValueError("execution timeout must be a zero or finite positive number.")
        self.exec_timeout = exec_timeout
        self.max_record_size = 10 * (2**20)  # 10 MBytes
        self.client_features = client_features or frozenset()
        if _zctx is None:
            _zctx = zmq.asyncio.Context()
        self.zctx = _zctx  # share the global context
        self._sockets = None
        self._is_socket_invalid = False
        self.completion_queue = asyncio.Queue(maxsize=128)
        self.service_queue = asyncio.Queue(maxsize=128)
        self.model_service_queue = asyncio.Queue(maxsize=128)
        self.service_apps_info_queue = asyncio.Queue(maxsize=128)
        self.status_queue = asyncio.Queue(maxsize=128)
        self.output_queue = None
        self.pending_queues = OrderedDict()
        self.current_run_id = None
        self.read_task = None
        self.status_task = None
        self.watchdog_task = None
        self._closed = False

    async def __ainit__(self) -> None:
        loop = current_loop()
        await self._get_socket_pair()
        self.status_task = loop.create_task(self.ping_status())
        self.read_task = loop.create_task(self.read_output())
        if self.exec_timeout > 0:
            self.watchdog_task = loop.create_task(self.watchdog())
        else:
            self.watchdog_task = None

    async def _create_sockets(self) -> SocketPair:
        input_sock = RobustSocket(zmq.PUSH, await self.get_repl_in_addr())
        output_sock = RobustSocket(zmq.PULL, await self.get_repl_out_addr())
        return SocketPair(input_sock, output_sock)

    async def _get_socket_pair(self) -> SocketPair:
        if self._sockets is None:
            self._sockets = await self._create_sockets()
        return self._sockets

    async def refresh_sockets(self) -> None:
        if self.read_task is not None:
            self.read_task.cancel()
        self._sockets = await self._create_sockets()
        loop = current_loop()
        self.read_task = loop.create_task(self.read_output())

    def __getstate__(self):
        props = self.__dict__.copy()
        del props["zctx"]
        del props["_sockets"]
        del props["_is_socket_invalid"]
        del props["completion_queue"]
        del props["service_queue"]
        del props["model_service_queue"]
        del props["service_apps_info_queue"]
        del props["status_queue"]
        del props["output_queue"]
        del props["pending_queues"]
        del props["read_task"]
        del props["status_task"]
        del props["watchdog_task"]
        del props["_closed"]
        del props["event_producer"]
        return props

    def __setstate__(self, props):
        global _zctx
        self.__dict__.update(props)
        if _zctx is None:
            _zctx = zmq.asyncio.Context()
        self.zctx = _zctx  # share the global context
        self._sockets = None
        self._is_socket_invalid = False
        self.completion_queue = asyncio.Queue(maxsize=128)
        self.service_queue = asyncio.Queue(maxsize=128)
        self.model_service_queue = asyncio.Queue(maxsize=128)
        self.service_apps_info_queue = asyncio.Queue(maxsize=128)
        self.status_queue = asyncio.Queue(maxsize=128)
        self.output_queue = None
        self.pending_queues = OrderedDict()
        self.read_task = None
        self.status_task = None
        self.watchdog_task = None
        self._closed = False
        # __ainit__() is called by the caller.

    @abstractmethod
    async def get_repl_in_addr(self) -> str:
        raise NotImplementedError

    @abstractmethod
    async def get_repl_out_addr(self) -> str:
        raise NotImplementedError

    async def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        try:
            if self.watchdog_task and not self.watchdog_task.done():
                self.watchdog_task.cancel()
                await self.watchdog_task
            if self.status_task and not self.status_task.done():
                self.status_task.cancel()
                await self.status_task
            if self.read_task and not self.read_task.done():
                self.read_task.cancel()
                await self.read_task
            if self._sockets is not None:
                self._sockets.close()
            # WARNING:
            # destroying zmq contexts here with possibility of re-entrance
            # may cause deadlocks.
        except Exception:
            log.exception("AbstractCodeRunner.close(): unexpected error")

    async def ping(self) -> dict[str, float] | None:
        try:
            return await self.feed_and_get_status()
        except Exception:
            log.exception("AbstractCodeRunner.ping(): unexpected error")
            return None

    async def ping_status(self):
        """
        This is to keep the REPL in/out port mapping in the Linux
        kernel's NAT table alive.
        """
        try:
            while True:
                ret = await self.feed_and_get_status()
                if ret is None:
                    break
                await asyncio.sleep(10)
        except asyncio.CancelledError:
            pass
        except Exception:
            log.exception("AbstractCodeRunner.ping_status(): unexpected error")

    async def feed_batch(self, opts):
        sock = await self._get_socket_pair()
        clean_cmd = opts.get("clean", "")
        if clean_cmd is None:
            clean_cmd = ""
        await sock.send_multipart([
            b"clean",
            clean_cmd.encode("utf8"),
        ])
        build_cmd = opts.get("build", "")
        if build_cmd is None:
            build_cmd = ""
        await sock.send_multipart([
            b"build",
            build_cmd.encode("utf8"),
        ])
        exec_cmd = opts.get("exec", "")
        if exec_cmd is None:
            exec_cmd = ""
        await sock.send_multipart([
            b"exec",
            exec_cmd.encode("utf8"),
        ])

    async def feed_code(self, text: str):
        sock = await self._get_socket_pair()
        await sock.send_multipart([b"code", text.encode("utf8")])

    async def feed_input(self, text: str):
        sock = await self._get_socket_pair()
        await sock.send_multipart([b"input", text.encode("utf8")])

    async def feed_event(self, evdata: AgentEventData):
        sock = await self._get_socket_pair()
        data = {
            "type": evdata.type,
            "data": evdata.data,
        }
        await sock.send_multipart([b"event", json.dumps(data).encode("utf8")])

    async def feed_interrupt(self):
        sock = await self._get_socket_pair()
        await sock.send_multipart([b"interrupt", b""])

    async def feed_and_get_status(self) -> dict[str, float] | None:
        sock = await self._get_socket_pair()
        await sock.send_multipart([b"status", b""])
        try:
            result = await self.status_queue.get()
            self.status_queue.task_done()
            return msgpack.unpackb(result)
        except asyncio.CancelledError:
            return None

    async def feed_and_get_completion(self, code_text, opts):
        sock = await self._get_socket_pair()
        payload = {
            "code": code_text,
        }
        payload.update(opts)
        await sock.send_multipart([
            b"complete",
            json.dumps(payload).encode("utf8"),
        ])
        try:
            result = await self.completion_queue.get()
            self.completion_queue.task_done()
            return json.loads(result)
        except asyncio.CancelledError:
            return []

    async def feed_start_model_service(self, model_info):
        sock = await self._get_socket_pair()
        await sock.send_multipart([
            b"start-model-service",
            json.dumps(model_info).encode("utf8"),
        ])
        if health_check_info := model_info.get("service", {}).get("health_check"):
            timeout_seconds = (
                health_check_info["max_retries"] * health_check_info["max_wait_time"] + 10
            )
        else:
            timeout_seconds = 10
        try:
            async with timeout(timeout_seconds):
                result = await self.model_service_queue.get()
            self.model_service_queue.task_done()
            return json.loads(result)
        except asyncio.CancelledError:
            return {"status": "failed", "error": "cancelled"}
        except asyncio.TimeoutError:
            return {"status": "failed", "error": "timeout"}

    async def feed_start_service(self, service_info):
        sock = await self._get_socket_pair()
        await sock.send_multipart([
            b"start-service",
            json.dumps(service_info).encode("utf8"),
        ])
        try:
            with timeout(10):
                result = await self.service_queue.get()
            self.service_queue.task_done()
            return json.loads(result)
        except asyncio.CancelledError:
            return {"status": "failed", "error": "cancelled"}
        except asyncio.TimeoutError:
            return {"status": "failed", "error": "timeout"}

    async def feed_shutdown_service(self, service_name: str):
        sock = await self._get_socket_pair()
        await sock.send_multipart([
            b"shutdown-service",
            json.dumps(service_name).encode("utf8"),
        ])

    async def feed_service_apps(self):
        sock = await self._get_socket_pair()
        await sock.send_multipart([
            b"get-apps",
            "".encode("utf8"),
        ])
        try:
            with timeout(10):
                result = await self.service_apps_info_queue.get()
            self.service_apps_info_queue.task_done()
            return json.loads(result)
        except asyncio.CancelledError:
            return {"status": "failed", "error": "cancelled"}
        except asyncio.TimeoutError:
            return {"status": "failed", "error": "timeout"}

    async def watchdog(self) -> None:
        try:
            await asyncio.sleep(self.exec_timeout)
            if self.output_queue is not None:
                # TODO: what to do if None?
                await self.output_queue.put(ResultRecord("exec-timeout", None))
        except asyncio.CancelledError:
            pass

    @staticmethod
    def aggregate_console(
        result: NextResult, records: Sequence[ResultRecord], api_ver: int
    ) -> None:
        if api_ver == 1:
            stdout_items = []
            stderr_items = []
            media_items = []
            html_items = []

            for rec in records:
                if rec.msg_type == "stdout":
                    stdout_items.append(rec.data or "")
                elif rec.msg_type == "stderr":
                    stderr_items.append(rec.data or "")
                elif rec.msg_type == "media" and rec.data is not None:
                    o = json.loads(rec.data)
                    media_items.append((o["type"], o["data"]))
                elif rec.msg_type == "html":
                    html_items.append(rec.data)

            result["stdout"] = "".join(stdout_items)
            result["stderr"] = "".join(stderr_items)
            result["media"] = media_items
            result["html"] = html_items

        elif api_ver >= 2:
            console_items: List[Tuple[ConsoleItemType, Union[str, Tuple[str, str]]]] = []
            last_stdout = io.StringIO()
            last_stderr = io.StringIO()

            for rec in records:
                if last_stdout.tell() and rec.msg_type != "stdout":
                    console_items.append(("stdout", last_stdout.getvalue()))
                    last_stdout.seek(0)
                    last_stdout.truncate(0)
                if last_stderr.tell() and rec.msg_type != "stderr":
                    console_items.append(("stderr", last_stderr.getvalue()))
                    last_stderr.seek(0)
                    last_stderr.truncate(0)

                if rec.msg_type == "stdout":
                    last_stdout.write(rec.data or "")
                elif rec.msg_type == "stderr":
                    last_stderr.write(rec.data or "")
                elif rec.msg_type == "media" and rec.data is not None:
                    o = json.loads(rec.data)
                    console_items.append(("media", (o["type"], o["data"])))
                elif rec.msg_type in outgoing_msg_types:
                    # FIXME: currently mypy cannot handle dynamic specialization of literals.
                    console_items.append((rec.msg_type, rec.data))  # type: ignore

            if last_stdout.tell():
                console_items.append(("stdout", last_stdout.getvalue()))
            if last_stderr.tell():
                console_items.append(("stderr", last_stderr.getvalue()))

            result["console"] = console_items
            last_stdout.close()
            last_stderr.close()

        else:
            raise AssertionError("Unrecognized API version")

    async def get_next_result(self, api_ver=2, flush_timeout=2.0) -> NextResult:
        # Context: per API request
        has_continuation = ClientFeatures.CONTINUATION in self.client_features
        records = []
        result: NextResult
        try:
            assert self.output_queue is not None
            with timeout(flush_timeout if has_continuation else None):
                while True:
                    rec = await self.output_queue.get()
                    if rec.msg_type in outgoing_msg_types:
                        records.append(rec)
                    self.output_queue.task_done()
                    if rec.msg_type == "finished":
                        data = json.loads(rec.data) if rec.data else {}
                        raise RunFinished(data)
                    elif rec.msg_type == "clean-finished":
                        data = json.loads(rec.data) if rec.data else {}
                        raise CleanFinished(data)
                    elif rec.msg_type == "build-finished":
                        data = json.loads(rec.data) if rec.data else {}
                        raise BuildFinished(data)
                    elif rec.msg_type == "waiting-input":
                        opts = json.loads(rec.data) if rec.data else {}
                        raise InputRequestPending(opts)
                    elif rec.msg_type == "exec-timeout":
                        raise ExecTimeout
        except asyncio.CancelledError:
            self.resume_output_queue()
            raise
        except asyncio.TimeoutError:
            result = {
                "runId": self.current_run_id,
                "status": "continued",
                "exitCode": None,
                "options": None,
            }
            type(self).aggregate_console(result, records, api_ver)
            self.resume_output_queue()
            return result
        except CleanFinished as e:
            result = {
                "runId": self.current_run_id,
                "status": "clean-finished",
                "exitCode": e.data.get("exitCode"),
                "options": None,
            }
            type(self).aggregate_console(result, records, api_ver)
            self.resume_output_queue()
            return result
        except BuildFinished as e:
            result = {
                "runId": self.current_run_id,
                "status": "build-finished",
                "exitCode": e.data.get("exitCode"),
                "options": None,
            }
            type(self).aggregate_console(result, records, api_ver)
            self.resume_output_queue()
            return result
        except RunFinished as e:
            result = {
                "runId": self.current_run_id,
                "status": "finished",
                "exitCode": e.data.get("exitCode"),
                "options": None,
            }
            type(self).aggregate_console(result, records, api_ver)
            self.next_output_queue()
            return result
        except ExecTimeout:
            result = {
                "runId": self.current_run_id,
                "status": "exec-timeout",
                "exitCode": None,
                "options": None,
            }
            log.warning(f"Execution timeout detected on kernel {self.kernel_id}")
            type(self).aggregate_console(result, records, api_ver)
            self.next_output_queue()
            return result
        except InputRequestPending as e:
            result = {
                "runId": self.current_run_id,
                "status": "waiting-input",
                "exitCode": None,
                "options": e.data,
            }
            type(self).aggregate_console(result, records, api_ver)
            self.resume_output_queue()
            return result
        except Exception:
            log.exception("unexpected error")
            raise

    async def attach_output_queue(self, run_id: Optional[str]) -> None:
        # Context: per API request
        if run_id is None:
            run_id = secrets.token_hex(16)
        assert run_id is not None
        if run_id not in self.pending_queues:
            q: asyncio.Queue[ResultRecord] = asyncio.Queue(maxsize=4096)
            activated = asyncio.Event()
            self.pending_queues[run_id] = (activated, q)
        else:
            activated, q = self.pending_queues[run_id]
        if self.output_queue is None:
            self.output_queue = q
        else:
            if self.current_run_id == run_id:
                # No need to wait if we are continuing.
                pass
            else:
                # If there is an outstanding ongoning execution,
                # wait until it has "finished".
                await activated.wait()
                activated.clear()
        self.current_run_id = run_id
        assert self.output_queue is q

    def resume_output_queue(self) -> None:
        """
        Use this to conclude get_next_result() when the execution should be
        continued from the client.

        At that time, we need to reuse the current run ID and its output queue.
        We don't change self.output_queue here so that we can continue to read
        outputs while the client sends the continuation request.
        """
        if self.current_run_id is None:
            return
        self.pending_queues.move_to_end(self.current_run_id, last=False)

    def next_output_queue(self) -> None:
        """
        Use this to conclude get_next_result() when we have finished a "run".
        """
        assert self.current_run_id is not None
        self.pending_queues.pop(self.current_run_id, None)
        self.current_run_id = None
        if len(self.pending_queues) > 0:
            # Make the next waiting API request handler to proceed.
            _, (activated, q) = self.pending_queues.popitem(last=False)
            self.output_queue = q
            activated.set()
        else:
            # If there is no pending request, just ignore all outputs
            # from the kernel.
            self.output_queue = None

    async def read_output(self) -> None:
        # We should use incremental decoder because some kernels may
        # send us incomplete UTF-8 byte sequences (e.g., Julia).
        decoders = (
            codecs.getincrementaldecoder("utf8")(errors="replace"),
            codecs.getincrementaldecoder("utf8")(errors="replace"),
        )
        sock = await self._get_socket_pair()
        while True:
            try:
                data = await sock.recv_multipart()
                if len(data) != 2:
                    log.warning(f"Invalid data from output socket, skip. (data: {data})")
                    continue
                msg_type, msg_data = data
                try:
                    match msg_type:
                        case b"status":
                            await self.status_queue.put(msg_data)
                        case b"completion":
                            await self.completion_queue.put(msg_data)
                        case b"service-result":
                            await self.service_queue.put(msg_data)
                        case b"model-service-result":
                            await self.model_service_queue.put(msg_data)
                        case b"model-service-status":
                            response = json.loads(msg_data)
                            event = ModelServiceStatusEvent(
                                self.kernel_id,
                                self.session_id,
                                response["model_name"],
                                (
                                    ModelServiceStatus.HEALTHY
                                    if response["is_healthy"]
                                    else ModelServiceStatus.UNHEALTHY
                                ),
                            )
                            await self.event_producer.produce_event(event)
                        case b"apps-result":
                            await self.service_apps_info_queue.put(msg_data)
                        case b"stdout":
                            if self.output_queue is None:
                                continue
                            if len(msg_data) > self.max_record_size:
                                msg_data = msg_data[: self.max_record_size]
                            await self.output_queue.put(
                                ResultRecord(
                                    "stdout",
                                    decoders[0].decode(msg_data),
                                )
                            )
                        case b"stderr":
                            if self.output_queue is None:
                                continue
                            if len(msg_data) > self.max_record_size:
                                msg_data = msg_data[: self.max_record_size]
                            await self.output_queue.put(
                                ResultRecord(
                                    "stderr",
                                    decoders[1].decode(msg_data),
                                )
                            )
                        case _:
                            # Normal outputs should go to the current
                            # output queue.
                            if self.output_queue is None:
                                continue
                            await self.output_queue.put(
                                ResultRecord(
                                    cast(ResultType, msg_type.decode("ascii")),
                                    msg_data.decode("utf8"),
                                )
                            )
                except asyncio.QueueFull:
                    pass
                if msg_type == b"build-finished":
                    # finalize incremental decoder
                    decoders[0].decode(b"", True)
                    decoders[1].decode(b"", True)
                elif msg_type == b"finished":
                    # finalize incremental decoder
                    decoders[0].decode(b"", True)
                    decoders[1].decode(b"", True)
                    self.finished_at = time.monotonic()
            except InvalidSocket:
                self._is_socket_invalid = True
                break
            except (asyncio.CancelledError, GeneratorExit):
                break
            except Exception:
                log.exception("unexpected error")
                break


def match_distro_data(data: Mapping[str, Any], distro: str) -> Tuple[str, Any]:
    """
    Find the latest or exactly matching entry from krunner_volumes mapping using the given distro
    string expression.

    It assumes that the keys of krunner_volumes mapping is a string concatenated with a distro
    prefix (e.g., "centos", "ubuntu") and a distro version composed of multiple integer components
    joined by single dots (e.g., "1.2.3", "18.04").
    """
    rx_ver_suffix = re.compile(r"(\d+(\.\d+)*)$")

    def _extract_version(key: str) -> Tuple[int, ...]:
        m = rx_ver_suffix.search(key)
        if m is not None:
            return tuple(map(int, m.group(1).split(".")))
        return (0,)

    m = rx_ver_suffix.search(distro)
    if m is None:
        # Assume latest
        distro_prefix = distro
        distro_ver = None
    else:
        distro_prefix = distro[: -len(m.group(1))]
        distro_ver = tuple(map(int, m.group(1).split(".")))

    # Check if there are static-build krunners first.
    if distro_prefix == "alpine":
        libc_flavor = "musl"
    else:
        libc_flavor = "gnu"
    distro_key = f"static-{libc_flavor}"
    if volume := data.get(distro_key):
        return distro_key, volume

    # Search through the per-distro versions
    match_list = [
        (distro_key, value, _extract_version(distro_key))
        for distro_key, value in data.items()
        if distro_key.startswith(distro_prefix)
    ]

    match_list = sorted(match_list, key=lambda x: x[2], reverse=True)
    if match_list:
        if distro_ver is None:
            return match_list[0][:-1]  # return latest
        for distro_key, value, matched_distro_ver in match_list:
            if distro_ver >= matched_distro_ver:
                return (distro_key, value)
        return match_list[-1][:-1]  # fallback to the latest of its kind
    raise UnsupportedBaseDistroError(distro)



from __future__ import annotations

import asyncio
import functools
import gzip
import io
import logging
import lzma
import os
import re
import shutil
import subprocess
import textwrap
from pathlib import Path, PurePosixPath
from typing import Any, Dict, Final, FrozenSet, Mapping, Optional, Sequence, Tuple, cast, override

import janus
import pkg_resources
from aiodocker.docker import Docker, DockerVolume
from aiodocker.exceptions import DockerError
from aiotools import TaskGroup

from ai.backend.agent.docker.utils import PersistentServiceContainer
from ai.backend.common.docker import ImageRef
from ai.backend.common.events import EventProducer
from ai.backend.common.lock import FileLock
from ai.backend.common.types import AgentId, CommitStatus, KernelId, Sentinel, SessionId
from ai.backend.common.utils import current_loop
from ai.backend.logging import BraceStyleAdapter
from ai.backend.plugin.entrypoint import scan_entrypoints

from ..kernel import AbstractCodeRunner, AbstractKernel
from ..resources import KernelResourceSpec
from ..types import AgentEventData
from ..utils import closing_async, get_arch_name

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

DEFAULT_CHUNK_SIZE: Final = 256 * 1024  # 256 KiB
DEFAULT_INFLIGHT_CHUNKS: Final = 8


class DockerKernel(AbstractKernel):
    network_driver: str

    def __init__(
        self,
        kernel_id: KernelId,
        session_id: SessionId,
        agent_id: AgentId,
        network_id: str,
        image: ImageRef,
        version: int,
        network_driver: str,
        *,
        agent_config: Mapping[str, Any],
        resource_spec: KernelResourceSpec,
        service_ports: Any,  # TODO: type-annotation
        environ: Mapping[str, Any],
        data: Dict[str, Any],
    ) -> None:
        super().__init__(
            kernel_id,
            session_id,
            agent_id,
            network_id,
            image,
            version,
            agent_config=agent_config,
            resource_spec=resource_spec,
            service_ports=service_ports,
            data=data,
            environ=environ,
        )

        self.network_driver = network_driver

    async def close(self) -> None:
        pass

    def __getstate__(self):
        props = super().__getstate__()
        return props

    def __setstate__(self, props):
        if "network_driver" not in props:
            props["network_driver"] = "bridge"
        super().__setstate__(props)

    async def create_code_runner(
        self, event_producer: EventProducer, *, client_features: FrozenSet[str], api_version: int
    ) -> AbstractCodeRunner:
        return await DockerCodeRunner.new(
            self.kernel_id,
            self.session_id,
            event_producer,
            kernel_host="127.0.0.1",  # repl ports are always bound to 127.0.0.1
            repl_in_port=self.data["repl_in_port"],
            repl_out_port=self.data["repl_out_port"],
            exec_timeout=0,
            client_features=client_features,
        )

    async def get_completions(self, text: str, opts: Mapping[str, Any]):
        assert self.runner is not None
        result = await self.runner.feed_and_get_completion(text, opts)
        return {"status": "finished", "completions": result}

    async def check_status(self):
        assert self.runner is not None
        result = await self.runner.feed_and_get_status()
        return result

    async def get_logs(self):
        container_id = self.data["container_id"]
        async with closing_async(Docker()) as docker:
            container = await docker.containers.get(container_id)
            logs = await container.log(stdout=True, stderr=True, follow=False)
        return {"logs": "".join(logs)}

    async def interrupt_kernel(self):
        assert self.runner is not None
        await self.runner.feed_interrupt()
        return {"status": "finished"}

    async def start_service(self, service: str, opts: Mapping[str, Any]):
        assert self.runner is not None
        if self.data.get("block_service_ports", False):
            return {
                "status": "failed",
                "error": "operation blocked",
            }
        for sport in self.service_ports:
            if sport["name"] == service:
                break
        else:
            return {"status": "failed", "error": "invalid service name"}
        result = await self.runner.feed_start_service({
            "name": service,
            "port": sport["container_ports"][0],  # primary port
            "ports": sport["container_ports"],
            "protocol": sport["protocol"],
            "options": opts,
        })
        return result

    async def start_model_service(self, model_service: Mapping[str, Any]):
        assert self.runner is not None
        result = await self.runner.feed_start_model_service(model_service)
        return result

    async def shutdown_service(self, service: str):
        assert self.runner is not None
        await self.runner.feed_shutdown_service(service)

    async def get_service_apps(self):
        assert self.runner is not None
        result = await self.runner.feed_service_apps()
        return result

    def _get_commit_path(self, kernel_id: KernelId, subdir: str) -> Tuple[Path, Path]:
        base_commit_path: Path = self.agent_config["agent"]["image-commit-path"]
        commit_path = base_commit_path / subdir
        lock_path = commit_path / "lock" / str(kernel_id)
        return commit_path, lock_path

    async def check_duplicate_commit(self, kernel_id: KernelId, subdir: str) -> CommitStatus:
        _, lock_path = self._get_commit_path(kernel_id, subdir)
        if lock_path.exists():
            return CommitStatus.ONGOING
        return CommitStatus.READY

    async def commit(
        self,
        kernel_id,
        subdir,
        *,
        canonical: str | None = None,
        filename: str | None = None,
        extra_labels: dict[str, str] = {},
    ) -> None:
        assert self.runner is not None

        loop = asyncio.get_running_loop()
        path, lock_path = self._get_commit_path(kernel_id, subdir)
        container_id: str = str(self.data["container_id"])
        try:
            Path(path).mkdir(exist_ok=True, parents=True)
            Path(lock_path).parent.mkdir(exist_ok=True, parents=True)
        except ValueError:  # parent_path does not start with work_dir!
            raise ValueError("malformed committed path.")

        def _write_chunks(
            fileobj: gzip.GzipFile,
            q: janus._SyncQueueProxy[bytes | Sentinel],
        ) -> None:
            while True:
                chunk = q.get()
                if chunk is Sentinel.TOKEN:
                    return
                fileobj.write(chunk)
                q.task_done()

        try:
            async with FileLock(path=lock_path, timeout=0.1, remove_when_unlock=True):
                log.info("Container (k: {}) is being committed", kernel_id)
                docker = Docker()
                try:
                    # There is a known issue at certain versions of Docker Engine
                    # which prevents container from being committed when request config body is empty
                    # https://github.com/moby/moby/issues/45543
                    docker_info = await docker.system.info()
                    docker_version = docker_info["ServerVersion"]
                    major, _, patch = docker_version.split(".", maxsplit=2)
                    config = None
                    if (int(major) == 23 and int(patch) < 8) or (
                        int(major) == 24 and int(patch) < 1
                    ):
                        config = {"ContainerSpec": {}}

                    container = docker.containers.container(container_id)
                    changes: list[str] = []

                    for label_name, label_value in extra_labels.items():
                        changes.append(f"LABEL {label_name}={label_value}")
                    if canonical:
                        if ":" in canonical:
                            repo, tag = canonical.rsplit(":", maxsplit=1)
                        else:
                            repo, tag = canonical, "latest"
                        log.debug("tagging image as {}:{}", repo, tag)
                    else:
                        repo, tag = None, None
                    response: Mapping[str, Any] = await container.commit(
                        changes=changes or None,
                        repository=repo,
                        tag=tag,
                        config=config,
                    )
                    image_id = response["Id"]
                    if filename:
                        filepath = path / filename
                        try:
                            q: janus.Queue[bytes | Sentinel] = janus.Queue(
                                maxsize=DEFAULT_INFLIGHT_CHUNKS
                            )
                            async with docker._query(f"images/{image_id}/get") as tb_resp:
                                with gzip.open(filepath, "wb") as fileobj:
                                    write_task = loop.run_in_executor(
                                        None,
                                        functools.partial(
                                            _write_chunks,
                                            fileobj,
                                            q.sync_q,
                                        ),
                                    )
                                    try:
                                        await asyncio.sleep(0)  # let write_task get started
                                        async for chunk in tb_resp.content.iter_chunked(
                                            DEFAULT_CHUNK_SIZE
                                        ):
                                            await q.async_q.put(chunk)
                                    finally:
                                        await q.async_q.put(Sentinel.TOKEN)
                                        await write_task
                        finally:
                            await docker.images.delete(image_id)
                finally:
                    await docker.close()
        except asyncio.TimeoutError:
            log.warning("Session is already being committed.")

    @override
    async def accept_file(self, container_path: os.PathLike | str, filedata: bytes) -> None:
        loop = current_loop()
        container_home_path = PurePosixPath("/home/work")
        try:
            home_relpath = PurePosixPath(container_path).relative_to(container_home_path)
        except ValueError:
            raise PermissionError("Not allowed to upload files outside /home/work")
        host_work_dir: Path = (
            self.agent_config["container"]["scratch-root"] / str(self.kernel_id) / "work"
        )
        host_abspath = (host_work_dir / home_relpath).resolve(strict=False)
        if not host_abspath.is_relative_to(host_work_dir):
            raise PermissionError("Not allowed to upload files outside /home/work")

        def _write_to_disk():
            host_abspath.parent.mkdir(parents=True, exist_ok=True)
            host_abspath.write_bytes(filedata)

        try:
            await loop.run_in_executor(None, _write_to_disk)
        except OSError as e:
            raise RuntimeError(
                "{0}: writing uploaded file failed: {1} -> {2} ({3})".format(
                    self.kernel_id,
                    container_path,
                    host_abspath,
                    repr(e),
                )
            )

    @override
    async def download_file(self, container_path: os.PathLike | str) -> bytes:
        container_id = self.data["container_id"]

        container_home_path = PurePosixPath("/home/work")
        container_abspath = PurePosixPath(os.path.normpath(container_home_path / container_path))
        if not container_abspath.is_relative_to(container_home_path):
            raise PermissionError("You cannot download files outside /home/work")

        async with closing_async(Docker()) as docker:
            container = docker.containers.container(container_id)
            try:
                with await container.get_archive(str(container_abspath)) as tarobj:
                    # FIXME: Replace this API call to a streaming version and cut the download if
                    #        the downloaded size exceeds the limit.
                    assert tarobj.fileobj is not None
                    tar_fobj = cast(io.BufferedIOBase, tarobj.fileobj)
                    tar_fobj.seek(0, io.SEEK_END)
                    tar_size = tar_fobj.tell()
                    if tar_size > 1048576:
                        raise ValueError("Too large archive file exceeding 1 MiB")
                    tar_fobj.seek(0, io.SEEK_SET)
                    tarbytes = tar_fobj.read()
            except DockerError:
                raise RuntimeError(f"Could not download the archive to: {container_abspath}")
        return tarbytes

    @override
    async def download_single(self, container_path: os.PathLike | str) -> bytes:
        container_id = self.data["container_id"]

        container_home_path = PurePosixPath("/home/work")
        container_abspath = PurePosixPath(os.path.normpath(container_home_path / container_path))
        if not container_abspath.is_relative_to(container_home_path):
            raise PermissionError("You cannot download files outside /home/work")

        async with closing_async(Docker()) as docker:
            container = docker.containers.container(container_id)
            try:
                with await container.get_archive(str(container_abspath)) as tarobj:
                    # FIXME: Replace this API call to a streaming version and cut the download if
                    #        the downloaded size exceeds the limit.
                    assert tarobj.fileobj is not None
                    tar_fobj = cast(io.BufferedIOBase, tarobj.fileobj)
                    tar_fobj.seek(0, io.SEEK_END)
                    tar_size = tar_fobj.tell()
                    if tar_size > 1048576:
                        raise ValueError("Too large archive file exceeding 1 MiB")
                    tar_fobj.seek(0, io.SEEK_SET)
                    if len(tarobj.getnames()) > 1:
                        raise ValueError(
                            f"Expected a single-file archive but found multiple files from {container_abspath}"
                        )
                    inner_fname = tarobj.getnames()[0]
                    inner_fobj = tarobj.extractfile(inner_fname)
                    if not inner_fobj:
                        raise ValueError(
                            f"Could not read {inner_fname!r} the archive file {container_abspath}"
                        )
                    # FYI: To get the size of extracted file, seek and tell with inner_fobj.
                    content_bytes = inner_fobj.read()
            except DockerError:
                raise RuntimeError(f"Could not download the archive to: {container_abspath}")
        return content_bytes

    @override
    async def list_files(self, container_path: os.PathLike | str):
        container_id = self.data["container_id"]

        # Confine the lookable paths in the home directory
        container_home_path = PurePosixPath("/home/work")
        container_abspath = PurePosixPath(os.path.normpath(container_home_path / container_path))
        if not container_abspath.is_relative_to(container_home_path):
            raise PermissionError("You cannot list files outside /home/work")

        # Gather individual file information in the target path.
        code = textwrap.dedent(
            """
        import json
        import os
        import stat
        import sys

        files = []
        for f in os.scandir(sys.argv[1]):
            fstat = f.stat(follow_symlinks=False)

            ctime = fstat.st_ctime  # TODO: way to get concrete create time?
            mtime = fstat.st_mtime
            atime = fstat.st_atime
            files.append({
                'mode': stat.filemode(fstat.st_mode),
                'size': fstat.st_size,
                'ctime': ctime,
                'mtime': mtime,
                'atime': atime,
                'filename': f.name,
            })
        print(json.dumps(files))
        """
        )
        proc = await asyncio.create_subprocess_exec(
            *[
                "docker",
                "exec",
                container_id,
                "/opt/backend.ai/bin/python",
                "-c",
                code,
                str(container_abspath),
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        raw_out, raw_err = await proc.communicate()
        out = raw_out.decode("utf-8")
        err = raw_err.decode("utf-8")
        return {"files": out, "errors": err, "abspath": str(container_path)}

    async def notify_event(self, evdata: AgentEventData):
        assert self.runner is not None
        await self.runner.feed_event(evdata)


class DockerCodeRunner(AbstractCodeRunner):
    kernel_host: str
    repl_in_port: int
    repl_out_port: int

    def __init__(
        self,
        kernel_id,
        session_id,
        event_producer,
        *,
        kernel_host,
        repl_in_port,
        repl_out_port,
        exec_timeout=0,
        client_features=None,
    ) -> None:
        super().__init__(
            kernel_id,
            session_id,
            event_producer,
            exec_timeout=exec_timeout,
            client_features=client_features,
        )
        self.kernel_host = kernel_host
        self.repl_in_port = repl_in_port
        self.repl_out_port = repl_out_port

    async def get_repl_in_addr(self) -> str:
        return f"tcp://{self.kernel_host}:{self.repl_in_port}"

    async def get_repl_out_addr(self) -> str:
        return f"tcp://{self.kernel_host}:{self.repl_out_port}"


async def prepare_krunner_env_impl(distro: str, entrypoint_name: str) -> Tuple[str, Optional[str]]:
    docker = Docker()
    arch = get_arch_name()
    current_version = int(
        Path(
            pkg_resources.resource_filename(
                f"ai.backend.krunner.{entrypoint_name}", f"./krunner-version.{distro}.txt"
            )
        )
        .read_text()
        .strip()
    )
    volume_name = f"backendai-krunner.v{current_version}.{arch}.{distro}"
    extractor_image = "backendai-krunner-extractor:latest"

    try:
        for item in await docker.images.list():
            if item["RepoTags"] is None or len(item["RepoTags"]) == 0:
                continue
            if item["RepoTags"][0] == extractor_image:
                break
        else:
            log.info("preparing the Docker image for krunner extractor...")
            extractor_archive = pkg_resources.resource_filename(
                "ai.backend.runner", f"krunner-extractor.img.{arch}.tar.xz"
            )
            with lzma.open(extractor_archive, "rb") as reader:
                image_tar = reader.read()
                proc = await asyncio.create_subprocess_exec(
                    *["docker", "load"], stdin=asyncio.subprocess.PIPE
                )
                await proc.communicate(input=image_tar)
                if proc.returncode != 0:
                    raise RuntimeError("loading krunner extractor image has failed!")

        log.info("checking krunner-env for {}...", distro)
        do_create = False
        try:
            vol = DockerVolume(docker, volume_name)
            await vol.show()
            # Instead of checking the version from txt files inside the volume,
            # we check the version via the volume name and its existence.
            # This is because:
            # - to avoid overwriting of volumes in use.
            # - the name comparison is quicker than reading actual files.
        except DockerError as e:
            if e.status == 404:
                do_create = True
        if do_create:
            archive_path = Path(
                pkg_resources.resource_filename(
                    f"ai.backend.krunner.{entrypoint_name}", f"krunner-env.{distro}.{arch}.tar.xz"
                )
            ).resolve()
            if not archive_path.exists():
                log.warning("krunner environment for {} ({}) is not supported!", distro, arch)
            else:
                log.info("populating {} volume version {}", volume_name, current_version)
                await docker.volumes.create({
                    "Name": volume_name,
                    "Driver": "local",
                })
                extractor_path = Path(
                    pkg_resources.resource_filename("ai.backend.runner", "krunner-extractor.sh")
                ).resolve()
                proc = await asyncio.create_subprocess_exec(*[
                    "docker",
                    "run",
                    "--rm",
                    "-i",
                    "-v",
                    f"{archive_path}:/root/archive.tar.xz",
                    "-v",
                    f"{extractor_path}:/root/krunner-extractor.sh",
                    "-v",
                    f"{volume_name}:/root/volume",
                    "-e",
                    f"KRUNNER_VERSION={current_version}",
                    extractor_image,
                    "/root/krunner-extractor.sh",
                ])
                if await proc.wait() != 0:
                    raise RuntimeError("extracting krunner environment has failed!")
    except Exception:
        log.exception("unexpected error")
        return distro, None
    finally:
        await docker.close()
    return distro, volume_name


async def prepare_krunner_env(local_config: Mapping[str, Any]) -> Mapping[str, Sequence[str]]:
    """
    Check if the volume "backendai-krunner.{distro}.{arch}" exists and is up-to-date.
    If not, automatically create it and update its content from the packaged pre-built krunner
    tar archives.
    """

    all_distros: list[tuple[str, str]] = []
    entry_prefix = "backendai_krunner_v10"
    for entrypoint in scan_entrypoints(entry_prefix):
        log.debug("loading krunner pkg: {}", entrypoint.module)
        plugin = entrypoint.load()
        await plugin.init({})  # currently does nothing
        provided_versions = (
            Path(
                pkg_resources.resource_filename(
                    f"ai.backend.krunner.{entrypoint.name}",
                    "versions.txt",
                )
            )
            .read_text()
            .splitlines()
        )
        all_distros.extend((distro, entrypoint.name) for distro in provided_versions)

    tasks = []
    async with TaskGroup() as tg:
        for distro, entrypoint_name in all_distros:
            tasks.append(tg.create_task(prepare_krunner_env_impl(distro, entrypoint_name)))
    distro_volumes = [t.result() for t in tasks if not t.cancelled()]
    result = {}
    for distro_name_and_version, volume_name in distro_volumes:
        if volume_name is None:
            continue
        result[distro_name_and_version] = volume_name
    return result


LinuxKit_IPTABLES_RULE = re.compile(
    r"DNAT\s+tcp\s+\-\-\s+anywhere\s+169\.254\.169\.254\s+tcp dpt:http to:127\.0\.0\.1:50128"
)
LinuxKit_CMD_EXEC_PREFIX = [
    "docker",
    "run",
    "--rm",
    "-i",
    "--privileged",
    "--pid=host",
    "linuxkit-nsenter:latest",
]


async def prepare_kernel_metadata_uri_handling(local_config: Mapping[str, Any]) -> None:
    if local_config["agent"]["docker-mode"] == "linuxkit":
        # Docker Desktop mode
        arch = get_arch_name()
        proxy_worker_binary = pkg_resources.resource_filename(
            "ai.backend.agent.docker", f"linuxkit-metadata-proxy-worker.{arch}.bin"
        )
        shutil.copyfile(proxy_worker_binary, "/tmp/backend.ai/linuxkit-metadata-proxy")
        os.chmod("/tmp/backend.ai/linuxkit-metadata-proxy", 0o755)
        server_port = local_config["agent"]["metadata-server-port"]
        # Prepare proxy worker container
        proxy_worker_container = PersistentServiceContainer(
            "linuxkit-nsenter:latest",
            {
                "Cmd": [
                    "/bin/sh",
                    "-c",
                    (
                        "ctr -n services.linuxkit t kill --exec-id metaproxy docker;ctr -n"
                        " services.linuxkit t exec --exec-id metaproxy docker"
                        " /host_mnt/tmp/backend.ai/linuxkit-metadata-proxy -remote-port"
                        f" {server_port}"
                    ),
                ],
                "HostConfig": {
                    "PidMode": "host",
                    "Privileged": True,
                },
            },
            name="linuxkit-nsenter",
        )
        await proxy_worker_container.ensure_running_latest()

        # Check if iptables rule is propagated on LinuxKit VM properly
        log.info("Checking metadata URL iptables rule ...")
        proc = await asyncio.create_subprocess_exec(
            *(LinuxKit_CMD_EXEC_PREFIX + ["/sbin/iptables", "-n", "-t", "nat", "-L", "PREROUTING"]),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        await proc.wait()
        assert proc.stdout is not None
        raw_rules = await proc.stdout.read()
        rules = raw_rules.decode()
        if LinuxKit_IPTABLES_RULE.search(rules) is None:
            proc = await asyncio.create_subprocess_exec(
                *(
                    LinuxKit_CMD_EXEC_PREFIX
                    + [
                        "/sbin/iptables",
                        "-t",
                        "nat",
                        "-I",
                        "PREROUTING",
                        "-d",
                        "169.254.169.254",
                        "-p",
                        "tcp",
                        "--dport",
                        "80",
                        "-j",
                        "DNAT",
                        "--to-destination",
                        "127.0.0.1:50128",
                    ]
                )
            )
            await proc.wait()
            log.info("Inserted the iptables rules.")
        else:
            log.info("The iptables rule already exists.")



import asyncio
import json
import logging
import os
import platform
from concurrent.futures import ProcessPoolExecutor
from decimal import Decimal
from pathlib import Path
from typing import (
    Any,
    Collection,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    cast,
)

import aiohttp
import async_timeout
import psutil
from aiodocker.docker import Docker, DockerContainer
from aiodocker.exceptions import DockerError

from ai.backend.agent.docker.kernel import DockerKernel
from ai.backend.agent.plugin.network import (
    AbstractNetworkAgentPlugin,
    ContainerNetworkCapability,
    ContainerNetworkInfo,
)
from ai.backend.agent.types import MountInfo
from ai.backend.common.netns import nsenter
from ai.backend.common.types import (
    AcceleratorMetadata,
    ClusterInfo,
    DeviceId,
    DeviceModelInfo,
    DeviceName,
    KernelCreationConfig,
    MetricKey,
    SlotName,
    SlotTypes,
)
from ai.backend.common.utils import current_loop, nmget
from ai.backend.logging import BraceStyleAdapter

from .. import __version__  # pants: no-infer-dep
from ..alloc_map import AllocationStrategy
from ..resources import (
    AbstractAllocMap,
    AbstractComputeDevice,
    AbstractComputePlugin,
    DeviceSlotInfo,
    DiscretePropertyAllocMap,
)
from ..stats import (
    ContainerMeasurement,
    Measurement,
    MetricTypes,
    NodeMeasurement,
    ProcessMeasurement,
    StatContext,
    StatModes,
)
from ..utils import closing_async, read_sysfs
from ..vendor.linux import libnuma
from .agent import Container
from .resources import get_resource_spec_from_container

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


# The list of pruned fstype when checking the filesystem usage statistics.
# Note that psutil's linux implementation automatically filters out "non-device" filesystems by
# checking /proc/filesystems so we don't have to put all the details virtual filesystems like
# "sockfs", "debugfs", etc.
pruned_disk_types = frozenset([
    "vfat",
    "lxcfs",
    "squashfs",
    "tmpfs",
    "iso9660",  # cdrom
])


def netstat_ns_work(ns_path: Path):
    with nsenter(ns_path):
        result = psutil.net_io_counters(pernic=True)
    return result


async def netstat_ns(ns_path: Path):
    loop = asyncio.get_running_loop()
    # Linux namespace is per-thread state. Therefore we need to ensure
    # IO is executed in the same thread where we switched the namespace.
    # Go provides runtime.LockOSThread() to do this.
    #
    # Unfortunately, CPython drops GIL while running IO and does not
    # provide any similar functionality. Therefore we execute namespace
    # dependent operation in the new process.
    with ProcessPoolExecutor(max_workers=1) as executor:
        result = await loop.run_in_executor(executor, netstat_ns_work, ns_path)
    return result


async def fetch_api_stats(container: DockerContainer) -> Optional[Dict[str, Any]]:
    short_cid = container._id[:7]
    try:
        ret = await container.stats(stream=False)  # TODO: cache
    except RuntimeError as e:
        msg = str(e.args[0]).lower()
        if "event loop is closed" in msg or "session is closed" in msg:
            return None
        raise
    except (DockerError, aiohttp.ClientError) as e:
        log.error(
            "cannot read stats (cid:{}): client error: {!r}.",
            short_cid,
            e,
        )
        return None
    else:
        entry = {"read": "0001-01-01"}
        # aiodocker 0.16 or later returns a list of dict, even when not streaming.
        match ret:
            case list() if ret:
                entry = ret[0]
            case dict() if ret:
                entry = ret
            case _:
                # The API may return an empty result upon container termination.
                log.warning(
                    "cannot read stats (cid:{}): got an empty result: {}",
                    short_cid,
                    ret,
                )
                return None
        if entry["read"].startswith("0001-01-01") or entry["preread"].startswith("0001-01-01"):
            return None
        return entry


# Pseudo-plugins for intrinsic devices (CPU and the main memory)


class CPUDevice(AbstractComputeDevice):
    pass


class CPUPlugin(AbstractComputePlugin):
    """
    Represents the CPU.
    """

    config_watch_enabled = False

    key = DeviceName("cpu")
    slot_types = [
        (SlotName("cpu"), SlotTypes.COUNT),
    ]

    async def init(self, context: Optional[Any] = None) -> None:
        pass

    async def cleanup(self) -> None:
        pass

    async def update_plugin_config(self, new_plugin_config: Mapping[str, Any]) -> None:
        pass

    async def list_devices(self) -> Collection[CPUDevice]:
        cores = await libnuma.get_available_cores()
        overcommit_factor = int(os.environ.get("BACKEND_CPU_OVERCOMMIT_FACTOR", "1"))
        assert 1 <= overcommit_factor <= 10
        return [
            CPUDevice(
                device_id=DeviceId(str(core_idx)),
                hw_location="root",
                numa_node=libnuma.node_of_cpu(core_idx),
                memory_size=0,
                processing_units=1 * overcommit_factor,
            )
            for core_idx in sorted(cores)
        ]

    async def available_slots(self) -> Mapping[SlotName, Decimal]:
        devices = await self.list_devices()
        return {
            SlotName("cpu"): Decimal(sum(dev.processing_units for dev in devices)),
        }

    def get_version(self) -> str:
        return __version__

    async def extra_info(self) -> Mapping[str, str]:
        return {
            "agent_version": __version__,
            "machine": platform.machine(),
            "os_type": platform.system(),
        }

    async def gather_node_measures(self, ctx: StatContext) -> Sequence[NodeMeasurement]:
        _cstat = psutil.cpu_times(True)
        q = Decimal("0.000")
        total_cpu_used = cast(
            Decimal, sum((Decimal(c.user + c.system) * 1000).quantize(q) for c in _cstat)
        )
        now, raw_interval = ctx.update_timestamp("cpu-node")
        interval = Decimal(raw_interval * 1000).quantize(q)

        return [
            NodeMeasurement(
                MetricKey("cpu_util"),
                MetricTypes.UTILIZATION,
                unit_hint="msec",
                current_hook=lambda metric: metric.stats.diff,
                per_node=Measurement(total_cpu_used, interval),
                per_device={
                    DeviceId(str(idx)): Measurement(
                        (Decimal(c.user + c.system) * 1000).quantize(q),
                        interval,
                    )
                    for idx, c in enumerate(_cstat)
                },
            ),
        ]

    async def gather_container_measures(
        self,
        ctx: StatContext,
        container_ids: Sequence[str],
    ) -> Sequence[ContainerMeasurement]:
        async def sysfs_impl(container_id):
            cpu_path = ctx.agent.get_cgroup_path("cpuacct", container_id)
            version = ctx.agent.docker_info["CgroupVersion"]
            try:
                match version:
                    case "1":
                        cpu_used = read_sysfs(cpu_path / "cpuacct.usage", int) / 1e6
                    case "2":
                        cpu_stats = {
                            k: v
                            for k, v in map(
                                lambda line: line.split(" "),
                                (cpu_path / "cpu.stat").read_text().splitlines(),
                            )
                        }
                        cpu_used = int(cpu_stats["usage_usec"]) / 1e3
            except IOError as e:
                log.warning(
                    "CPUPlugin: cannot read stats: sysfs unreadable for container {0}\n{1!r}",
                    container_id[:7],
                    e,
                )
                return None
            return cpu_used

        async def api_impl(container_id):
            async with closing_async(Docker()) as docker:
                container = DockerContainer(docker, id=container_id)
                try:
                    async with async_timeout.timeout(2.0):
                        ret = await fetch_api_stats(container)
                except asyncio.TimeoutError:
                    return None
                if ret is None:
                    return None
                cpu_used = nmget(ret, "cpu_stats.cpu_usage.total_usage", 0) / 1e6
                return cpu_used

        if ctx.mode == StatModes.CGROUP:
            impl = sysfs_impl
        elif ctx.mode == StatModes.DOCKER:
            impl = api_impl
        else:
            raise RuntimeError("should not reach here")

        q = Decimal("0.000")
        per_container_cpu_used = {}
        per_container_cpu_util = {}
        tasks = []
        for cid in container_ids:
            tasks.append(asyncio.create_task(impl(cid)))
        results = await asyncio.gather(*tasks)
        for cid, cpu_used in zip(container_ids, results):
            if cpu_used is None:
                continue
            per_container_cpu_used[cid] = Measurement(Decimal(cpu_used).quantize(q))
            per_container_cpu_util[cid] = Measurement(
                Decimal(cpu_used).quantize(q),
                capacity=Decimal(1000),
            )
        return [
            ContainerMeasurement(
                MetricKey("cpu_util"),
                MetricTypes.UTILIZATION,
                unit_hint="percent",
                current_hook=lambda metric: metric.stats.rate,
                stats_filter=frozenset({"avg", "max"}),
                per_container=per_container_cpu_util,
            ),
            ContainerMeasurement(
                MetricKey("cpu_used"),
                MetricTypes.ACCUMULATION,
                unit_hint="msec",
                per_container=per_container_cpu_used,
            ),
        ]

    async def gather_process_measures(
        self, ctx: StatContext, pid_map: Mapping[int, str]
    ) -> Sequence[ProcessMeasurement]:
        async def psutil_impl(pid: int) -> Optional[Decimal]:
            try:
                p = psutil.Process(pid)
            except psutil.NoSuchProcess:
                log.warning("psutil cannot found process {0}", pid)
            else:
                cpu_times = p.cpu_times()
                cpu_used = Decimal(cpu_times.user + cpu_times.system) * 1000
                return cpu_used
            return None

        async def api_impl(cid: str, pids: List[int]) -> List[Optional[Decimal]]:
            return []

        per_process_cpu_util = {}
        per_process_cpu_used = {}
        results: List[Decimal | None] = []
        q = Decimal("0.000")
        pid_map_list = list(pid_map.items())
        match self.local_config["agent"]["docker-mode"]:
            case "linuxkit":
                api_tasks: list[asyncio.Task[list[Decimal | None]]] = []
                # group by container ID
                cid_pids_map: Dict[str, List[int]] = {}
                for pid, cid in pid_map_list:
                    if cid_pids_map.get(cid) is None:
                        cid_pids_map[cid] = []
                    cid_pids_map[cid].append(pid)
                for cid, pids in cid_pids_map.items():
                    api_tasks.append(asyncio.create_task(api_impl(cid, pids)))
                chunked_results = await asyncio.gather(*api_tasks)
                for chunk in chunked_results:
                    results.extend(chunk)
            case _:
                psutil_tasks = []
                for pid, _ in pid_map_list:
                    psutil_tasks.append(asyncio.create_task(psutil_impl(pid)))
                results = await asyncio.gather(*psutil_tasks)

        for (pid, cid), cpu_used in zip(pid_map_list, results):
            if cpu_used is None:
                continue
            per_process_cpu_util[pid] = Measurement(
                Decimal(cpu_used).quantize(q), capacity=Decimal(1000)
            )
            per_process_cpu_used[pid] = Measurement(Decimal(cpu_used).quantize(q))
        return [
            ProcessMeasurement(
                MetricKey("cpu_util"),
                MetricTypes.UTILIZATION,
                unit_hint="percent",
                current_hook=lambda metric: metric.stats.rate,
                stats_filter=frozenset({"avg", "max"}),
                per_process=per_process_cpu_util,
            ),
            ProcessMeasurement(
                MetricKey("cpu_used"),
                MetricTypes.ACCUMULATION,
                unit_hint="msec",
                per_process=per_process_cpu_used,
            ),
        ]

    async def create_alloc_map(self) -> AbstractAllocMap:
        devices = await self.list_devices()
        return DiscretePropertyAllocMap(
            device_slots={
                dev.device_id: DeviceSlotInfo(
                    SlotTypes.COUNT, SlotName("cpu"), Decimal(dev.processing_units)
                )
                for dev in devices
            },
        )

    async def get_hooks(self, distro: str, arch: str) -> Sequence[Path]:
        # TODO: move the sysconf hook in libbaihook.so here
        return []

    async def generate_docker_args(
        self,
        docker: Docker,
        device_alloc,
    ) -> Mapping[str, Any]:
        cores = [*map(int, device_alloc["cpu"].keys())]
        sorted_core_ids = [*map(str, sorted(cores))]
        return {
            "HostConfig": {
                "Cpus": len(cores),
                "CpusetCpus": ",".join(sorted_core_ids),
                # 'CpusetMems': f'{resource_spec.numa_node}',
            },
        }

    async def restore_from_container(
        self,
        container: Container,
        alloc_map: AbstractAllocMap,
    ) -> None:
        assert isinstance(alloc_map, DiscretePropertyAllocMap)
        # Docker does not return the original cpuset.... :(
        # We need to read our own records.
        resource_spec = await get_resource_spec_from_container(container.backend_obj)
        if resource_spec is None:
            return
        alloc_map.apply_allocation({
            SlotName("cpu"): resource_spec.allocations[DeviceName("cpu")][SlotName("cpu")],
        })

    async def get_attached_devices(
        self,
        device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]],
    ) -> Sequence[DeviceModelInfo]:
        device_ids = [*device_alloc[SlotName("cpu")].keys()]
        available_devices = await self.list_devices()
        attached_devices: List[DeviceModelInfo] = []
        for device in available_devices:
            if device.device_id in device_ids:
                attached_devices.append({
                    "device_id": device.device_id,
                    "model_name": "",
                    "data": {"cores": len(device_ids)},
                })
        return attached_devices

    async def get_docker_networks(
        self, device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]]
    ) -> List[str]:
        return []

    async def generate_mounts(
        self, source_path: Path, device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]]
    ) -> List[MountInfo]:
        return []

    def get_metadata(self) -> AcceleratorMetadata:
        return {
            "slot_name": "cpu",
            "description": "CPU",
            "human_readable_name": "CPU",
            "display_unit": "Core",
            "number_format": {"binary": False, "round_length": 0},
            "display_icon": "cpu",
        }


class MemoryDevice(AbstractComputeDevice):
    pass


class MemoryPlugin(AbstractComputePlugin):
    """
    Represents the main memory.

    When collecting statistics, it also measures network and I/O usage
    in addition to the memory usage.
    """

    config_watch_enabled = False

    key = DeviceName("mem")
    slot_types = [
        (SlotName("mem"), SlotTypes.BYTES),
    ]

    async def init(self, context: Optional[Any] = None) -> None:
        pass

    async def cleanup(self) -> None:
        pass

    async def update_plugin_config(self, new_plugin_config: Mapping[str, Any]) -> None:
        pass

    async def list_devices(self) -> Collection[MemoryDevice]:
        memory_size = psutil.virtual_memory().total
        overcommit_factor = int(os.environ.get("BACKEND_MEM_OVERCOMMIT_FACTOR", "1"))
        return [
            MemoryDevice(
                device_id=DeviceId("root"),
                device_name=self.key,
                hw_location="root",
                numa_node=0,  # the kernel setting will do the job.
                memory_size=overcommit_factor * memory_size,
                processing_units=0,
            ),
        ]

    async def available_slots(self) -> Mapping[SlotName, Decimal]:
        devices = await self.list_devices()
        return {
            SlotName("mem"): Decimal(sum(dev.memory_size for dev in devices)),
        }

    def get_version(self) -> str:
        return __version__

    async def extra_info(self) -> Mapping[str, str]:
        return {}

    async def gather_node_measures(self, ctx: StatContext) -> Sequence[NodeMeasurement]:
        _mstat = psutil.virtual_memory()
        total_mem_used_bytes = Decimal(_mstat.total - _mstat.available)
        total_mem_capacity_bytes = Decimal(_mstat.total)
        _nstat = psutil.net_io_counters()
        net_rx_bytes = _nstat.bytes_recv
        net_tx_bytes = _nstat.bytes_sent

        def get_disk_stat():
            total_disk_usage = Decimal(0)
            total_disk_capacity = Decimal(0)
            per_disk_stat = {}
            for disk_info in psutil.disk_partitions():
                # Skip additional filesystem types not filtered by psutil, like squashfs.
                if disk_info.fstype in pruned_disk_types:
                    continue
                # Skip transient filesystems created/destroyed by Docker.
                if disk_info.mountpoint.startswith("/proc/docker/runtime-runc/moby/"):
                    continue
                # Skip btrfs subvolumes used by Docker if configured.
                if disk_info.mountpoint == "/var/lib/docker/btrfs":
                    continue
                dstat = os.statvfs(disk_info.mountpoint)
                disk_usage = Decimal(dstat.f_frsize * (dstat.f_blocks - dstat.f_bavail))
                disk_capacity = Decimal(dstat.f_frsize * dstat.f_blocks)
                per_disk_stat[disk_info.device] = Measurement(disk_usage, disk_capacity)
                total_disk_usage += disk_usage
                total_disk_capacity += disk_capacity
            return total_disk_usage, total_disk_capacity, per_disk_stat

        loop = current_loop()
        total_disk_usage, total_disk_capacity, per_disk_stat = await loop.run_in_executor(
            None, get_disk_stat
        )
        return [
            NodeMeasurement(
                MetricKey("mem"),
                MetricTypes.GAUGE,
                unit_hint="bytes",
                stats_filter=frozenset({"max"}),
                per_node=Measurement(total_mem_used_bytes, total_mem_capacity_bytes),
                per_device={
                    DeviceId("root"): Measurement(total_mem_used_bytes, total_mem_capacity_bytes)
                },
            ),
            NodeMeasurement(
                MetricKey("disk"),
                MetricTypes.GAUGE,
                unit_hint="bytes",
                per_node=Measurement(total_disk_usage, total_disk_capacity),
                per_device=per_disk_stat,
            ),
            NodeMeasurement(
                MetricKey("net_rx"),
                MetricTypes.RATE,
                unit_hint="bps",
                current_hook=lambda metric: metric.stats.rate,
                per_node=Measurement(Decimal(net_rx_bytes)),
                per_device={DeviceId("node"): Measurement(Decimal(net_rx_bytes))},
            ),
            NodeMeasurement(
                MetricKey("net_tx"),
                MetricTypes.RATE,
                unit_hint="bps",
                current_hook=lambda metric: metric.stats.rate,
                per_node=Measurement(Decimal(net_tx_bytes)),
                per_device={DeviceId("node"): Measurement(Decimal(net_tx_bytes))},
            ),
        ]

    async def gather_container_measures(
        self, ctx: StatContext, container_ids: Sequence[str]
    ) -> Sequence[ContainerMeasurement]:
        def get_scratch_size(container_id: str) -> int:
            # Temporarily disabled as this function incurs too much delay with
            # a large number of files in scratch dirs, causing indefinite accumulation of
            # stat collector tasks and slowing down everything.
            return 0
            # for kernel_id, info in ctx.agent.kernel_registry.items():
            #     if info['container_id'] == container_id:
            #         break
            # else:
            #     return 0
            # work_dir = ctx.agent.local_config['container']['scratch-root'] / str(kernel_id) / 'work'
            # total_size = 0
            # for path in work_dir.rglob('*'):
            #     if path.is_symlink():
            #         total_size += path.lstat().st_size
            #     elif path.is_file():
            #         total_size += path.stat().st_size
            # return total_size

        async def sysfs_impl(container_id):
            mem_path = ctx.agent.get_cgroup_path("memory", container_id)
            io_path = ctx.agent.get_cgroup_path("blkio", container_id)
            version = ctx.agent.docker_info["CgroupVersion"]

            try:
                io_read_bytes = 0
                io_write_bytes = 0
                match version:
                    case "1":
                        mem_cur_bytes = read_sysfs(mem_path / "memory.usage_in_bytes", int)
                        mem_max_bytes = read_sysfs(mem_path / "memory.limit_in_bytes", int)

                        for line in (mem_path / "memory.stat").read_text().splitlines():
                            key, value = line.split(" ")
                            if key == "total_inactive_file":
                                mem_cur_bytes -= int(value)
                                break

                        # example data:
                        #   8:0 Read 13918208
                        #   8:0 Write 0
                        #   8:0 Sync 0
                        #   8:0 Async 13918208
                        #   8:0 Total 13918208
                        #   Total 13918208
                        for line in (
                            (io_path / "blkio.throttle.io_service_bytes").read_text().splitlines()
                        ):
                            if line.startswith("Total "):
                                continue
                            dev, op, nbytes = line.strip().split()
                            if op == "Read":
                                io_read_bytes += int(nbytes)
                            elif op == "Write":
                                io_write_bytes += int(nbytes)
                    case "2":
                        mem_cur_bytes = read_sysfs(mem_path / "memory.current", int)
                        mem_max_bytes = read_sysfs(mem_path / "memory.max", int)

                        for line in (mem_path / "memory.stat").read_text().splitlines():
                            key, value = line.split(" ")
                            if key == "inactive_file":
                                mem_cur_bytes -= int(value)
                                break

                        # example data:
                        # 8:16 rbytes=1459200 wbytes=314773504 rios=192 wios=353 dbytes=0 dios=0
                        # 8:0 rbytes=3387392 wbytes=176128 rios=103 wios=32 dbytes=0 dios=0
                        for line in (io_path / "io.stat").read_text().splitlines():
                            for io_stat in line.split()[1:]:
                                stat, value = io_stat.split("=")
                                if stat == "rbytes":
                                    io_read_bytes += int(value)
                                if stat == "wbytes":
                                    io_write_bytes += int(value)
            except IOError as e:
                log.warning(
                    "MemoryPlugin: cannot read stats: sysfs unreadable for container {0}\n{1!r}",
                    container_id[:7],
                    e,
                )
                return None
            async with closing_async(Docker()) as docker:
                container = DockerContainer(docker, id=container_id)
                data = await container.show()
                sandbox_key = data["NetworkSettings"]["SandboxKey"]
            net_rx_bytes = 0
            net_tx_bytes = 0
            nstat = await netstat_ns(sandbox_key)
            for name, stat in nstat.items():
                if name == "lo":
                    continue
                net_rx_bytes += stat.bytes_recv
                net_tx_bytes += stat.bytes_sent
            loop = current_loop()
            scratch_sz = await loop.run_in_executor(None, get_scratch_size, container_id)
            return (
                mem_cur_bytes,
                mem_max_bytes,
                io_read_bytes,
                io_write_bytes,
                net_rx_bytes,
                net_tx_bytes,
                scratch_sz,
            )

        async def api_impl(container_id):
            async with closing_async(Docker()) as docker:
                container = DockerContainer(docker, id=container_id)
                try:
                    async with async_timeout.timeout(2.0):
                        ret = await fetch_api_stats(container)
                except asyncio.TimeoutError:
                    return None
                if ret is None:
                    return None
                mem_cur_bytes = nmget(ret, "memory_stats.usage", 0)
                mem_total_bytes = nmget(ret, "memory_stats.limit", 0)
                io_read_bytes = 0
                io_write_bytes = 0
                for item in nmget(ret, "blkio_stats.io_service_bytes_recursive", []):
                    if item["op"] == "Read":
                        io_read_bytes += item["value"]
                    elif item["op"] == "Write":
                        io_write_bytes += item["value"]
                net_rx_bytes = 0
                net_tx_bytes = 0
                for name, stat in ret["networks"].items():
                    net_rx_bytes += stat["rx_bytes"]
                    net_tx_bytes += stat["tx_bytes"]
                loop = current_loop()
                scratch_sz = await loop.run_in_executor(None, get_scratch_size, container_id)
                return (
                    mem_cur_bytes,
                    mem_total_bytes,
                    io_read_bytes,
                    io_write_bytes,
                    net_rx_bytes,
                    net_tx_bytes,
                    scratch_sz,
                )

        if ctx.mode == StatModes.CGROUP:
            impl = sysfs_impl
        elif ctx.mode == StatModes.DOCKER:
            impl = api_impl
        else:
            raise RuntimeError("should not reach here")

        per_container_mem_used_bytes = {}
        per_container_io_read_bytes = {}
        per_container_io_write_bytes = {}
        per_container_net_rx_bytes = {}
        per_container_net_tx_bytes = {}
        per_container_io_scratch_size = {}
        tasks = []
        for cid in container_ids:
            tasks.append(asyncio.create_task(impl(cid)))
        results = await asyncio.gather(*tasks)
        for cid, result in zip(container_ids, results):
            if result is None:
                continue
            per_container_mem_used_bytes[cid] = Measurement(
                Decimal(result[0]), capacity=Decimal(result[1])
            )
            per_container_io_read_bytes[cid] = Measurement(Decimal(result[2]))
            per_container_io_write_bytes[cid] = Measurement(Decimal(result[3]))
            per_container_net_rx_bytes[cid] = Measurement(Decimal(result[4]))
            per_container_net_tx_bytes[cid] = Measurement(Decimal(result[5]))
            per_container_io_scratch_size[cid] = Measurement(Decimal(result[6]))
        return [
            ContainerMeasurement(
                MetricKey("mem"),
                MetricTypes.GAUGE,
                unit_hint="bytes",
                stats_filter=frozenset({"max"}),
                per_container=per_container_mem_used_bytes,
            ),
            ContainerMeasurement(
                MetricKey("io_read"),
                MetricTypes.GAUGE,
                unit_hint="bytes",
                stats_filter=frozenset({"rate"}),
                per_container=per_container_io_read_bytes,
            ),
            ContainerMeasurement(
                MetricKey("io_write"),
                MetricTypes.GAUGE,
                unit_hint="bytes",
                stats_filter=frozenset({"rate"}),
                per_container=per_container_io_write_bytes,
            ),
            ContainerMeasurement(
                MetricKey("net_rx"),
                MetricTypes.RATE,
                unit_hint="bps",
                current_hook=lambda metric: metric.stats.rate,
                per_container=per_container_net_rx_bytes,
            ),
            ContainerMeasurement(
                MetricKey("net_tx"),
                MetricTypes.RATE,
                unit_hint="bps",
                current_hook=lambda metric: metric.stats.rate,
                per_container=per_container_net_tx_bytes,
            ),
            ContainerMeasurement(
                MetricKey("io_scratch_size"),
                MetricTypes.GAUGE,
                unit_hint="bytes",
                stats_filter=frozenset({"max"}),
                per_container=per_container_io_scratch_size,
            ),
        ]

    async def gather_process_measures(
        self, ctx: StatContext, pid_map: Mapping[int, str]
    ) -> Sequence[ProcessMeasurement]:
        async def psutil_impl(pid) -> Tuple[Optional[int], Optional[int], Optional[int]]:
            try:
                p = psutil.Process(pid)
            except psutil.NoSuchProcess:
                log.warning("psutil cannot found process {0}", pid)
            else:
                stats = p.as_dict(attrs=["memory_info", "io_counters"])
                mem_cur_bytes = io_read_bytes = io_write_bytes = None
                if stats["memory_info"] is not None:
                    mem_cur_bytes = stats["memory_info"].rss
                if stats["io_counters"] is not None:
                    io_read_bytes = stats["io_counters"].read_bytes
                    io_write_bytes = stats["io_counters"].write_bytes
                return mem_cur_bytes, io_read_bytes, io_write_bytes
            return None, None, None

        async def api_impl(
            cid: str, pids: List[int]
        ) -> List[Tuple[Optional[int], Optional[int], Optional[int]]]:
            return []

        per_process_mem_used_bytes = {}
        per_process_io_read_bytes = {}
        per_process_io_write_bytes = {}
        results: List[Tuple[Optional[int], Optional[int], Optional[int]]]
        pid_map_list = list(pid_map.items())
        match self.local_config["agent"]["docker-mode"]:
            case "linuxkit":
                api_tasks = []
                # group by container ID
                cid_pids_map: Dict[str, List[int]] = {}
                for pid, cid in pid_map_list:
                    if cid_pids_map.get(cid) is None:
                        cid_pids_map[cid] = []
                    cid_pids_map[cid].append(pid)
                for cid, pids in cid_pids_map.items():
                    api_tasks.append(asyncio.create_task(api_impl(cid, pids)))
                chunked_results = await asyncio.gather(*api_tasks)
                results = []
                for chunk in chunked_results:
                    results.extend(chunk)
            case _:
                psutil_tasks = []
                for pid, _ in pid_map_list:
                    psutil_tasks.append(asyncio.create_task(psutil_impl(pid)))
                results = await asyncio.gather(*psutil_tasks)

        for (pid, _), result in zip(pid_map_list, results):
            mem, io_read, io_write = result
            if mem is not None:
                per_process_mem_used_bytes[pid] = Measurement(Decimal(mem))
            if io_read is not None:
                per_process_io_read_bytes[pid] = Measurement(Decimal(io_read))
            if io_write is not None:
                per_process_io_write_bytes[pid] = Measurement(Decimal(io_write))
        return [
            ProcessMeasurement(
                MetricKey("mem"),
                MetricTypes.GAUGE,
                unit_hint="bytes",
                stats_filter=frozenset({"max"}),
                per_process=per_process_mem_used_bytes,
            ),
            ProcessMeasurement(
                MetricKey("io_read"),
                MetricTypes.GAUGE,
                unit_hint="bytes",
                stats_filter=frozenset({"rate"}),
                per_process=per_process_io_read_bytes,
            ),
            ProcessMeasurement(
                MetricKey("io_write"),
                MetricTypes.GAUGE,
                unit_hint="bytes",
                stats_filter=frozenset({"rate"}),
                per_process=per_process_io_write_bytes,
            ),
        ]

    async def create_alloc_map(self) -> AbstractAllocMap:
        devices = await self.list_devices()
        return DiscretePropertyAllocMap(
            allocation_strategy=AllocationStrategy.FILL,
            device_slots={
                dev.device_id: DeviceSlotInfo(
                    SlotTypes.BYTES, SlotName("mem"), Decimal(dev.memory_size)
                )
                for dev in devices
            },
        )

    async def get_hooks(self, distro: str, arch: str) -> Sequence[Path]:
        return []

    async def generate_docker_args(
        self,
        docker: Docker,
        device_alloc,
    ) -> Mapping[str, Any]:
        memory = sum(device_alloc["mem"].values())
        return {
            "HostConfig": {
                "MemorySwap": int(memory),  # prevent using swap!
                "Memory": int(memory),
            },
        }

    async def restore_from_container(
        self,
        container: Container,
        alloc_map: AbstractAllocMap,
    ) -> None:
        assert isinstance(alloc_map, DiscretePropertyAllocMap)
        memory_limit = container.backend_obj["HostConfig"]["Memory"]
        alloc_map.apply_allocation({
            SlotName("mem"): {DeviceId("root"): memory_limit},
        })

    async def get_attached_devices(
        self,
        device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]],
    ) -> Sequence[DeviceModelInfo]:
        device_ids = [*device_alloc[SlotName("mem")].keys()]
        available_devices = await self.list_devices()
        attached_devices: List[DeviceModelInfo] = []
        for device in available_devices:
            if device.device_id in device_ids:
                attached_devices.append({
                    "device_id": device.device_id,
                    "model_name": "",
                    "data": {},
                })
        return attached_devices

    async def get_docker_networks(
        self, device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]]
    ) -> List[str]:
        return []

    async def generate_mounts(
        self, source_path: Path, device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]]
    ) -> List[MountInfo]:
        return []

    def get_metadata(self) -> AcceleratorMetadata:
        return {
            "slot_name": "ram",
            "description": "Memory",
            "human_readable_name": "RAM",
            "display_unit": "GiB",
            "number_format": {"binary": True, "round_length": 0},
            "display_icon": "ram",
        }


class OverlayNetworkPlugin(AbstractNetworkAgentPlugin[DockerKernel]):
    async def init(self, context: Any = None) -> None:
        pass

    async def cleanup(self) -> None:
        pass

    async def update_plugin_config(self, plugin_config: Mapping[str, Any]) -> None:
        return await super().update_plugin_config(plugin_config)

    async def get_capabilities(self) -> Set[ContainerNetworkCapability]:
        return set()

    async def join_network(
        self,
        kernel_config: KernelCreationConfig,
        cluster_info: ClusterInfo,
        **kwargs: Any,
    ) -> dict[str, Any]:
        network_name: str = kwargs["network_name"]
        return {
            "HostConfig": {
                "NetworkMode": network_name,
            },
            "NetworkingConfig": {
                "EndpointsConfig": {
                    network_name: {
                        "Aliases": [kernel_config["cluster_hostname"]],
                    },
                },
            },
        }

    async def leave_network(self, kernel: DockerKernel) -> None:
        pass


class HostNetworkPlugin(AbstractNetworkAgentPlugin[DockerKernel]):
    async def init(self, context: Any = None) -> None:
        pass

    async def cleanup(self) -> None:
        pass

    async def update_plugin_config(self, plugin_config: Mapping[str, Any]) -> None:
        return await super().update_plugin_config(plugin_config)

    async def get_capabilities(self) -> Set[ContainerNetworkCapability]:
        return set([ContainerNetworkCapability.GLOBAL])

    async def join_network(
        self, kernel_config: KernelCreationConfig, cluster_info: ClusterInfo, **kwargs
    ) -> Dict[str, Any]:
        if _cluster_ssh_port_mapping := cluster_info.get("cluster_ssh_port_mapping"):
            return {
                "HostConfig": {
                    "ExtraHosts": [
                        f"{hostname}:{host_port[0]}"
                        for hostname, host_port in _cluster_ssh_port_mapping.items()
                    ],
                    "NetworkMode": "host",
                }
            }
        else:
            return {
                "HostConfig": {
                    "NetworkMode": "host",
                },
            }

    async def leave_network(self, kernel: DockerKernel) -> None:
        pass

    async def prepare_port_forward(
        self, kernel: DockerKernel, bind_host: str, ports: Iterable[Tuple[int, int]], **kwargs
    ) -> None:
        host_ports = [p[0] for p in ports]
        scratch_dir = (
            self.local_config["container"]["scratch-root"] / str(kernel.kernel_id)
        ).resolve()
        config_dir = scratch_dir / "config"

        intrinsic_ports = {
            "replin": host_ports[0],
            "replout": host_ports[1],
        }
        for index, port_info in enumerate(kernel.service_ports):
            port_name = port_info["name"]
            if port_name in ("sshd", "ttyd"):
                intrinsic_ports[port_name] = host_ports[index + 2]

        await current_loop().run_in_executor(
            None,
            lambda: (config_dir / "intrinsic-ports.json").write_text(json.dumps(intrinsic_ports)),
        )

    async def expose_ports(
        self, kernel: DockerKernel, bind_host: str, ports: Iterable[Tuple[int, int]], **kwargs
    ) -> ContainerNetworkInfo:
        host_ports = [p[0] for p in ports]

        intrinsic_ports = {
            "replin": host_ports[0],
            "replout": host_ports[1],
        }
        for index, port_info in enumerate(kernel.service_ports):
            port_name = port_info["name"]
            if port_name in ("sshd", "ttyd"):
                intrinsic_ports[port_name] = host_ports[index + 2]

        return ContainerNetworkInfo(
            bind_host,
            {service_name: {port: port} for service_name, port in intrinsic_ports.items()},
        )
